"""A central location to define QTrio specific exceptions and avoid introducing
inter-module dependency issues."""
import typing
from qtpy import QtCore


class QTrioException(Exception):
    """Base exception for all QTrio exceptions."""

    # https://github.com/sphinx-doc/sphinx/issues/7493
    __module__ = "qtrio"


class NoOutcomesError(QTrioException):
    """Raised if you try to unwrap a :class:`qtrio.Outcomes` which has no outcomes."""

    # https://github.com/sphinx-doc/sphinx/issues/7493
    __module__ = "qtrio"


class EventTypeRegistrationError(QTrioException):
    """Base class for various event type registration exceptions to inherit from."""

    # https://github.com/sphinx-doc/sphinx/issues/7493
    __module__ = "qtrio"


class EventTypeRegistrationFailedError(EventTypeRegistrationError):
    """Raised if the attempt to register a new event type fails."""

    # https://github.com/sphinx-doc/sphinx/issues/7493
    __module__ = "qtrio"

    def __init__(self) -> None:
        super().__init__(
            "Failed to register the event hint, either no available hints remain or the"
            + " program is shutting down."
        )


class RequestedEventTypeUnavailableError(EventTypeRegistrationError):
    """Raised if the requested event type is unavailable."""

    # https://github.com/sphinx-doc/sphinx/issues/7493
    __module__ = "qtrio"

    def __init__(
        self,
        requested_type: typing.Union[int, QtCore.QEvent.Type],
        returned_type: typing.Union[int, QtCore.QEvent.Type],
    ) -> None:
        super().__init__(
            f"Failed acquire the requested type ({requested_type}), got back"
            + f" ({returned_type}) instead."
        )


class EventTypeAlreadyRegisteredError(EventTypeRegistrationError):
    """Raised when a request is made to register an event type but a type has already
    been registered previously."""

    # https://github.com/sphinx-doc/sphinx/issues/7493
    __module__ = "qtrio"

    def __init__(self) -> None:
        super().__init__(
            "An event type has already been registered, this must only happen once."
        )


class ReturnCodeError(QTrioException):
    """Wraps a QApplication return code as an exception."""

    # https://github.com/sphinx-doc/sphinx/issues/7493
    __module__ = "qtrio"

    def __eq__(self, other: object) -> bool:
        if type(self) != type(other):
            return False

        # TODO: workaround for https://github.com/python/mypy/issues/4445
        if not isinstance(other, type(self)):  # pragma: no cover
            return False

        return self.args == other.args


class InternalError(QTrioException):
    """Raised when an internal state is inconsistent."""

    # https://github.com/sphinx-doc/sphinx/issues/7493
    __module__ = "qtrio"


class UserCancelledError(QTrioException):
    """Raised when a user requested cancellation of an operation."""

    # https://github.com/sphinx-doc/sphinx/issues/7493
    __module__ = "qtrio"


class RunnerTimedOutError(QTrioException):
    """Raised when a :class:`qtrio.Runner` times out the run."""

    # https://github.com/sphinx-doc/sphinx/issues/7493
    __module__ = "qtrio"


class InvalidInputError(QTrioException):
    """Raised when invalid input is provided such as via a dialog."""

    # https://github.com/sphinx-doc/sphinx/issues/7493
    __module__ = "qtrio"


class DialogNotActiveError(QTrioException):
    """Raised when attempting to interact with a dialog while it is not actually
    available.
    """

    # https://github.com/sphinx-doc/sphinx/issues/7493
    __module__ = "qtrio"



import contextlib
import os
import sys
import typing

import async_generator
import attr
from qtpy import QtWidgets
import trio

import qtrio
import qtrio._qt
import qtrio._util


if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol


class BasicDialogProtocol(Protocol):
    """The minimal common interface used for working with QTrio dialogs.  To check that
    a class implements this protocol see
    :func:`qtrio.dialogs.check_basic_dialog_protocol`.
    """

    finished: qtrio.Signal
    """The signal to be emitted when the dialog is finished."""

    def setup(self) -> None:
        """Setup and show the dialog.  Emit :attr:`qtrio.dialogs.DialogProtocol.shown`
        when done.
        """

    def teardown(self) -> None:
        """Hide and teardown the dialog."""


check_basic_dialog_protocol = qtrio._util.ProtocolChecker[BasicDialogProtocol]()
"""Assert proper implementation of :class:`qtrio.dialogs.BasicDialogProtocol` when type
hint checking.

Arguments:
    cls: The class to verify.

Returns:
    The same class, unmodified, that was passed in.
"""


class DialogProtocol(BasicDialogProtocol, Protocol):
    """The common interface used for working with QTrio dialogs.  To check that a class
    implements this protocol see :func:`qtrio.dialogs.check_dialog_protocol`.
    """

    shown: qtrio.Signal
    """The signal to be emitted when the dialog is shown."""

    async def wait(self) -> object:
        """Show the dialog, wait for the user interaction, and return the result.

        Raises:
            qtrio.InvalidInputError: If the input can't be parsed as an integer.
            qtrio.UserCancelledError: If the user cancels the dialog.
        """


check_dialog_protocol = qtrio._util.ProtocolChecker[DialogProtocol]()
"""Assert proper implementation of :class:`qtrio.dialogs.DialogProtocol` when type hint
checking.

Arguments:
    cls: The class to verify.

Returns:
    The same class, unmodified, that was passed in.
"""


@contextlib.contextmanager
def _manage(dialog: BasicDialogProtocol) -> typing.Generator[trio.Event, None, None]:
    """Manage the setup and teardown of a dialog including yielding a
    :class:`trio.Event` that is set when the dialog is finished.

    Arguments:
        dialog: The dialog to be managed.
    """
    finished_event = trio.Event()

    def slot(*args: object, **kwargs: object) -> None:
        """Accept and ignore all arguments, then set the event."""
        finished_event.set()

    with qtrio._qt.connection(signal=dialog.finished, slot=slot):
        try:
            dialog.setup()
            yield finished_event
        finally:
            dialog.teardown()


def _dialog_button_box_buttons_by_role(
    dialog: QtWidgets.QDialog,
) -> typing.Mapping[QtWidgets.QDialogButtonBox.ButtonRole, QtWidgets.QAbstractButton]:
    """Create a mapping from button roles to their corresponding buttons."""

    hits = dialog.findChildren(QtWidgets.QDialogButtonBox)

    if len(hits) == 0:
        return {}

    [button_box] = hits
    return {button_box.buttonRole(button): button for button in button_box.buttons()}


@check_dialog_protocol
@attr.s(auto_attribs=True)
class IntegerDialog:
    """Manage a dialog for inputting an integer from the user.  Generally instances
    should be built via :func:`qtrio.dialogs.create_integer_dialog`."""

    parent: typing.Optional[QtWidgets.QWidget] = None
    """The parent widget for the dialog."""

    dialog: typing.Optional[QtWidgets.QInputDialog] = None
    """The actual dialog widget instance."""
    edit_widget: typing.Optional[QtWidgets.QLineEdit] = None
    """The line edit that the user will enter the input into."""
    accept_button: typing.Optional[QtWidgets.QPushButton] = None
    """The entry confirmation button."""
    reject_button: typing.Optional[QtWidgets.QPushButton] = None
    """The input cancellation button."""

    result: typing.Optional[int] = None
    """The result of parsing the user input."""

    shown = qtrio.Signal(QtWidgets.QInputDialog)
    """See :attr:`qtrio.dialogs.DialogProtocol.shown`."""
    finished = qtrio.Signal(int)  # QtWidgets.QDialog.DialogCode
    """See :attr:`qtrio.dialogs.BasicDialogProtocol.finished`."""

    def setup(self) -> None:
        """See :meth:`qtrio.dialogs.BasicDialogProtocol.setup`."""

        self.result = None

        self.dialog = QtWidgets.QInputDialog(self.parent)

        # TODO: adjust so we can use a context manager?
        self.dialog.finished.connect(self.finished)

        self.dialog.show()

        buttons = _dialog_button_box_buttons_by_role(dialog=self.dialog)
        self.accept_button = buttons.get(QtWidgets.QDialogButtonBox.AcceptRole)
        self.reject_button = buttons.get(QtWidgets.QDialogButtonBox.RejectRole)

        [self.edit_widget] = self.dialog.findChildren(QtWidgets.QLineEdit)

        self.shown.emit(self.dialog)

    def teardown(self) -> None:
        """See :meth:`qtrio.dialogs.BasicDialogProtocol.teardown`."""

        if self.dialog is not None:
            self.dialog.close()
            self.dialog.finished.disconnect(self.finished)
        self.dialog = None
        self.accept_button = None
        self.reject_button = None
        self.edit_widget = None

    async def wait(self) -> int:
        """See :meth:`qtrio.dialogs.DialogProtocol.wait`."""
        with _manage(dialog=self) as finished_event:
            if self.dialog is None:  # pragma: no cover
                raise qtrio.InternalError(
                    "Dialog not assigned while it is being managed."
                )

            await finished_event.wait()

            if self.dialog.result() != QtWidgets.QDialog.Accepted:
                raise qtrio.UserCancelledError()

            try:
                self.result = int(self.dialog.textValue())
            except ValueError:
                raise qtrio.InvalidInputError()

            return self.result


def create_integer_dialog(
    parent: typing.Optional[QtWidgets.QWidget] = None,
) -> IntegerDialog:
    """Create an integer input dialog.

    Arguments:
        parent: See :attr:`qtrio.dialogs.IntegerDialog.parent`.

    Returns:
        The dialog manager.
    """
    return IntegerDialog(parent=parent)


@check_dialog_protocol
@attr.s(auto_attribs=True)
class TextInputDialog:
    """Manage a dialog for inputting an integer from the user.  Generally instances
    should be built via :func:`qtrio.dialogs.create_text_input_dialog`."""

    title: typing.Optional[str] = None
    """The title of the dialog."""
    label: typing.Optional[str] = None
    """The label for the input widget."""
    parent: typing.Optional[QtWidgets.QWidget] = None
    """The parent widget for the dialog."""

    dialog: typing.Optional[QtWidgets.QInputDialog] = None
    """The actual dialog widget instance."""
    accept_button: typing.Optional[QtWidgets.QPushButton] = None
    """The entry confirmation button."""
    reject_button: typing.Optional[QtWidgets.QPushButton] = None
    """The input cancellation button."""
    line_edit: typing.Optional[QtWidgets.QLineEdit] = None
    """The line edit that the user will enter the input into."""

    result: typing.Optional[str] = None
    """The result of parsing the user input."""

    shown = qtrio.Signal(QtWidgets.QInputDialog)
    """See :attr:`qtrio.dialogs.DialogProtocol.shown`."""
    finished = qtrio.Signal(int)  # QtWidgets.QDialog.DialogCode
    """See :attr:`qtrio.dialogs.BasicDialogProtocol.finished`."""

    def setup(self) -> None:
        """See :meth:`qtrio.dialogs.BasicDialogProtocol.setup`."""

        self.result = None

        self.dialog = QtWidgets.QInputDialog(parent=self.parent)
        if self.label is not None:
            self.dialog.setLabelText(self.label)
        if self.title is not None:
            self.dialog.setWindowTitle(self.title)

        # TODO: adjust so we can use a context manager?
        self.dialog.finished.connect(self.finished)
        self.dialog.show()

        buttons = _dialog_button_box_buttons_by_role(dialog=self.dialog)
        self.accept_button = buttons[QtWidgets.QDialogButtonBox.AcceptRole]
        self.reject_button = buttons[QtWidgets.QDialogButtonBox.RejectRole]

        [self.line_edit] = self.dialog.findChildren(QtWidgets.QLineEdit)

        self.shown.emit(self.dialog)

    def teardown(self) -> None:
        """See :meth:`qtrio.dialogs.BasicDialogProtocol.teardown`."""

        if self.dialog is not None:
            self.dialog.close()
            self.dialog.finished.disconnect(self.finished)
        self.dialog = None
        self.accept_button = None
        self.reject_button = None

    async def wait(self) -> str:
        """See :meth:`qtrio.dialogs.DialogProtocol.wait`."""

        with _manage(dialog=self) as finished_event:
            if self.dialog is None:  # pragma: no cover
                raise qtrio.InternalError(
                    "Dialog not assigned while it is being managed."
                )

            await finished_event.wait()

            dialog_result = self.dialog.result()

            if dialog_result == QtWidgets.QDialog.Rejected:
                raise qtrio.UserCancelledError()

            # TODO: `: str` is a workaround for
            #       https://github.com/spyder-ide/qtpy/pull/217
            text_result: str = self.dialog.textValue()

            self.result = text_result

            return text_result


def create_text_input_dialog(
    title: typing.Optional[str] = None,
    label: typing.Optional[str] = None,
    parent: typing.Optional[QtWidgets.QWidget] = None,
) -> TextInputDialog:
    """Create a text input dialog.

    Arguments:
        title: The text to use for the input dialog title bar.
        label: The text to use for the input text box label.
        parent: See :attr:`qtrio.dialogs.IntegerDialog.parent`.

    Returns:
        The dialog manager.
    """
    return TextInputDialog(title=title, label=label, parent=parent)


@check_dialog_protocol
@attr.s(auto_attribs=True)
class FileDialog:
    """Manage a dialog for allowing the user to select a file or directory.  Generally
    instances should be built via :func:`qtrio.dialogs.create_file_save_dialog`."""

    file_mode: QtWidgets.QFileDialog.FileMode
    """Controls whether the dialog is for picking an existing vs. new file or directory,
    etc.
    """
    accept_mode: QtWidgets.QFileDialog.AcceptMode
    """Specify an open vs. a save dialog."""
    default_directory: typing.Optional[trio.Path] = None
    """The directory to be initially presented in the dialog."""
    default_file: typing.Optional[trio.Path] = None
    """The file to be initially selected in the dialog."""
    options: QtWidgets.QFileDialog.Option = QtWidgets.QFileDialog.Option()
    """Miscellaneous options.  See the Qt documentation."""
    parent: typing.Optional[QtWidgets.QWidget] = None
    """The parent widget for the dialog."""

    dialog: typing.Optional[QtWidgets.QFileDialog] = None
    """The actual dialog widget instance."""
    accept_button: typing.Optional[QtWidgets.QPushButton] = None
    """The confirmation button."""
    reject_button: typing.Optional[QtWidgets.QPushButton] = None
    """The cancellation button."""
    file_name_line_edit: typing.Optional[QtWidgets.QLineEdit] = None
    """The file name line edit widget."""

    result: typing.Optional[trio.Path] = None
    """The path selected by the user."""

    shown = qtrio.Signal(QtWidgets.QFileDialog)
    """See :attr:`qtrio.dialogs.DialogProtocol.shown`."""
    finished = qtrio.Signal(int)  # QtWidgets.QDialog.DialogCode
    """See :attr:`qtrio.dialogs.BasicDialogProtocol.finished`."""

    async def set_path(self, path: trio.Path) -> None:
        """Set the directory and enter the file name in the text box.  Note that this
        does not select the file in the file list.

        Arguments:
            path: The full path to the file to be set.
        """
        if self.dialog is None:
            raise qtrio.DialogNotActiveError(
                "File dialog not available for interaction."
            )

        self.dialog.setDirectory(os.fspath(path.parent))

        if self.file_name_line_edit is None:  # pragma: no cover
            raise qtrio.InternalError("File name line edit unexpectedly not known.")

        self.file_name_line_edit.setText(path.name)

    def setup(self) -> None:
        """See :meth:`qtrio.dialogs.BasicDialogProtocol.setup`."""

        self.result = None

        extras = {}

        if self.default_directory is not None:
            extras["directory"] = os.fspath(self.default_directory)

        options = self.options
        if sys.platform == "darwin":
            # https://github.com/altendky/qtrio/issues/28
            options |= QtWidgets.QFileDialog.DontUseNativeDialog

        self.dialog = QtWidgets.QFileDialog(
            parent=self.parent, options=options, **extras
        )

        if self.default_file is not None:
            self.dialog.selectFile(os.fspath(self.default_file))

        self.dialog.setFileMode(self.file_mode)
        self.dialog.setAcceptMode(self.accept_mode)

        # TODO: adjust so we can use a context manager?
        self.dialog.finished.connect(self.finished)

        self.dialog.show()

        buttons = _dialog_button_box_buttons_by_role(dialog=self.dialog)
        self.accept_button = buttons.get(QtWidgets.QDialogButtonBox.AcceptRole)
        self.reject_button = buttons.get(QtWidgets.QDialogButtonBox.RejectRole)
        [self.file_name_line_edit] = self.dialog.findChildren(QtWidgets.QLineEdit)

        self.shown.emit(self.dialog)

    def teardown(self) -> None:
        """See :meth:`qtrio.dialogs.BasicDialogProtocol.teardown`."""

        if self.dialog is not None:
            self.dialog.close()
            self.dialog.finished.disconnect(self.finished)
        self.dialog = None
        self.accept_button = None
        self.reject_button = None
        self.file_name_line_edit = None

    async def wait(self) -> trio.Path:
        """See :meth:`qtrio.dialogs.DialogProtocol.wait`."""

        with _manage(dialog=self) as finished_event:
            if self.dialog is None:  # pragma: no cover
                raise qtrio.InternalError(
                    "Dialog not assigned while it is being managed."
                )

            await finished_event.wait()
            if self.dialog.result() != QtWidgets.QDialog.Accepted:
                raise qtrio.UserCancelledError()

            [path_string] = self.dialog.selectedFiles()
            self.result = trio.Path(path_string)

            return self.result


def create_file_save_dialog(
    parent: typing.Optional[QtWidgets.QWidget] = None,
    default_directory: typing.Optional[trio.Path] = None,
    default_file: typing.Optional[trio.Path] = None,
    options: QtWidgets.QFileDialog.Option = QtWidgets.QFileDialog.Option(),
) -> FileDialog:
    """Create a file save dialog.

    Arguments:
        parent: See :attr:`qtrio.dialogs.FileDialog.parent`.
        default_directory: See :attr:`qtrio.dialogs.FileDialog.default_directory`.
        default_file: See :attr:`qtrio.dialogs.FileDialog.default_file`.
        options: See :attr:`qtrio.dialogs.FileDialog.options`.
    """
    return FileDialog(
        parent=parent,
        default_directory=default_directory,
        default_file=default_file,
        options=options,
        file_mode=QtWidgets.QFileDialog.AnyFile,
        accept_mode=QtWidgets.QFileDialog.AcceptSave,
    )


def create_file_open_dialog(
    parent: typing.Optional[QtWidgets.QWidget] = None,
    default_directory: typing.Optional[trio.Path] = None,
    default_file: typing.Optional[trio.Path] = None,
    options: QtWidgets.QFileDialog.Option = QtWidgets.QFileDialog.Option(),
) -> FileDialog:
    """Create a file open dialog.

    Arguments:
        parent: See :attr:`qtrio.dialogs.FileDialog.parent`.
        default_directory: See :attr:`qtrio.dialogs.FileDialog.default_directory`.
        default_file: See :attr:`qtrio.dialogs.FileDialog.default_file`.
        options: See :attr:`qtrio.dialogs.FileDialog.options`.
    """
    return FileDialog(
        parent=parent,
        default_directory=default_directory,
        default_file=default_file,
        options=options,
        file_mode=QtWidgets.QFileDialog.AnyFile,
        accept_mode=QtWidgets.QFileDialog.AcceptOpen,
    )


@check_dialog_protocol
@attr.s(auto_attribs=True)
class MessageBox:
    """Manage a message box for notifying the user.  Generally instances should be built
    via :func:`qtrio.dialogs.create_message_box`.
    """

    title: str
    """The message box title."""
    text: str
    """The message text shown inside the dialog."""
    icon: QtWidgets.QMessageBox.Icon
    """The icon shown inside the dialog."""
    buttons: QtWidgets.QMessageBox.StandardButtons
    """The buttons to be shown in the dialog."""
    parent: typing.Optional[QtWidgets.QWidget] = None
    """The parent widget for the dialog."""

    dialog: typing.Optional[QtWidgets.QMessageBox] = None
    """The actual dialog widget instance."""
    accept_button: typing.Optional[QtWidgets.QPushButton] = None
    """The button to accept the dialog."""

    result: typing.Optional[trio.Path] = None
    """Not generally relevant for a message box."""

    shown = qtrio.Signal(QtWidgets.QMessageBox)
    """See :attr:`qtrio.dialogs.DialogProtocol.shown`."""
    finished = qtrio.Signal(int)  # QtWidgets.QDialog.DialogCode
    """See :attr:`qtrio.dialogs.BasicDialogProtocol.finished`."""

    def setup(self) -> None:
        """See :meth:`qtrio.dialogs.BasicDialogProtocol.setup`."""

        self.result = None

        self.dialog = QtWidgets.QMessageBox(
            self.icon, self.title, self.text, self.buttons, self.parent
        )

        # TODO: adjust so we can use a context manager?
        self.dialog.finished.connect(self.finished)

        self.dialog.show()

        buttons = _dialog_button_box_buttons_by_role(dialog=self.dialog)
        self.accept_button = buttons[QtWidgets.QDialogButtonBox.AcceptRole]

        self.shown.emit(self.dialog)

    def teardown(self) -> None:
        """See :meth:`qtrio.dialogs.BasicDialogProtocol.teardown`."""

        if self.dialog is not None:
            self.dialog.close()
        self.dialog = None
        self.accept_button = None

    async def wait(self) -> None:
        """See :meth:`qtrio.dialogs.DialogProtocol.wait`."""

        with _manage(dialog=self) as finished_event:
            if self.dialog is None:  # pragma: no cover
                raise qtrio.InternalError(
                    "Dialog not assigned while it is being managed."
                )

            await finished_event.wait()

            result = self.dialog.result()

            if result == QtWidgets.QDialog.Rejected:
                raise qtrio.UserCancelledError()


def create_message_box(
    title: str = "",
    text: str = "",
    icon: QtWidgets.QMessageBox.Icon = QtWidgets.QMessageBox.Information,
    buttons: QtWidgets.QMessageBox.StandardButtons = QtWidgets.QMessageBox.Ok,
    parent: typing.Optional[QtWidgets.QWidget] = None,
) -> MessageBox:
    """Create a message box.

    Arguments:
        title: See :attr:`qtrio.dialogs.MessageBox.title`.
        text: See :attr:`qtrio.dialogs.MessageBox.text`.
        icon: See :attr:`qtrio.dialogs.MessageBox.icon`.
        buttons: See :attr:`qtrio.dialogs.MessageBox.buttons`.
        parent: See :attr:`qtrio.dialogs.MessageBox.parent`.
    """
    return MessageBox(icon=icon, title=title, text=text, buttons=buttons, parent=parent)


@check_basic_dialog_protocol
@attr.s(auto_attribs=True)
class ProgressDialog:
    """Manage a progress dialog for updating the user.  Generally instances should be
    built via :func:`qtrio.dialogs.create_progress_dialog`.
    """

    title: str
    """The message box title."""
    text: str
    """The message text shown as the progress bar label."""
    cancel_button_text: typing.Optional[str]
    """The cancel button text."""
    minimum: int
    """The progress value corresponding to no progress."""
    maximum: int
    """The progress value corresponding to completion."""

    parent: typing.Optional[QtWidgets.QWidget] = None
    """The parent widget for the dialog."""

    dialog: typing.Optional[QtWidgets.QProgressDialog] = None
    """The actual dialog widget instance."""
    cancel_button: typing.Optional[QtWidgets.QPushButton] = None
    """The cancellation button."""

    shown = qtrio.Signal(QtWidgets.QMessageBox)
    """See :attr:`qtrio.dialogs.DialogProtocol.shown`."""
    finished = qtrio.Signal(int)  # QtWidgets.QDialog.DialogCode
    """See :attr:`qtrio.dialogs.BasicDialogProtocol.finished`."""

    def setup(self) -> None:
        """See :meth:`qtrio.dialogs.BasicDialogProtocol.setup`."""

        self.final_value = None

        self.dialog = QtWidgets.QProgressDialog()
        self.dialog.setWindowTitle(self.title)
        self.dialog.setLabelText(self.text)
        self.dialog.setMinimum(self.minimum)
        self.dialog.setMaximum(self.maximum)
        self.dialog.setParent(self.parent)

        # TODO: adjust so we can use a context manager?
        self.dialog.finished.connect(self.finished)

        if self.cancel_button_text is not None:
            self.dialog.setCancelButtonText(self.cancel_button_text)

        self.dialog.show()

        [self.cancel_button] = self.dialog.findChildren(QtWidgets.QPushButton)

        self.shown.emit(self.dialog)

    def teardown(self) -> None:
        """See :meth:`qtrio.dialogs.BasicDialogProtocol.teardown`."""

        if self.dialog is not None:
            self.dialog.close()
            self.dialog.finished.disconnect(self.finished)
        self.dialog = None
        self.cancel_button = None

    @async_generator.asynccontextmanager
    async def manage(self) -> typing.AsyncIterator[None]:
        """A context manager to setup the progress dialog, cancel the managed context
        and teardown the dialog when done.
        """
        with _manage(dialog=self):
            if self.dialog is None:  # pragma: no cover
                raise qtrio.InternalError(
                    "Dialog not assigned while it is being managed."
                )

            with trio.CancelScope() as cancel_scope:
                with qtrio._qt.connection(
                    signal=self.dialog.canceled, slot=cancel_scope.cancel
                ):
                    yield

            if self.dialog.wasCanceled():
                raise qtrio.UserCancelledError()


def create_progress_dialog(
    title: str = "",
    text: str = "",
    cancel_button_text: typing.Optional[str] = None,
    minimum: int = 0,
    maximum: int = 0,
    parent: typing.Optional[QtWidgets.QWidget] = None,
) -> ProgressDialog:
    """Create a progress dialog.

    Arguments:
        title: See :attr:`qtrio.dialogs.ProgressDialog.title`.
        text: See :attr:`qtrio.dialogs.ProgressDialog.text`.
        cancel_button_text: See :attr:`qtrio.dialogs.ProgressDialog.cancel_button_text`.
        minimum: See :attr:`qtrio.dialogs.ProgressDialog.minimum`.
        maximum: See :attr:`qtrio.dialogs.ProgressDialog.maximum`.
        parent: See :attr:`qtrio.dialogs.ProgressDialog.parent`.
    """
    return ProgressDialog(
        title=title,
        text=text,
        cancel_button_text=cancel_button_text,
        minimum=minimum,
        maximum=maximum,
        parent=parent,
    )



import pytest


timeout = 40


def test_assertion_failure_fails(testdir):
    """QTrio hosted test which fails an assertion fails the test."""

    test_file = r"""
    import trio

    async def test():
        await trio.sleep(0)
        assert False
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(failed=1)


def test_assertion_pass_passes(testdir):
    """QTrio hosted test which passes an assertion passes the test."""

    test_file = r"""
    import trio

    async def test():
        await trio.sleep(0)
        assert True
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)



import typing

import attr
from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import qtrio


class QSignaledWidget(QtWidgets.QWidget):
    """A :class:`QtWidgets.QWidget` with extra signals for events of interest.

    Attributes:
        closed: A signal that will be emitted after a close event.
    """

    closed = QtCore.Signal()
    shown = QtCore.Signal()

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        """Detect close events and emit the ``closed`` signal."""

        super().closeEvent(event)
        if event.isAccepted():
            self.closed.emit()
        else:  # pragma: no cover
            pass

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        """Detect show events and emit the ``shown`` signal."""

        super().showEvent(event)
        if event.isAccepted():
            self.shown.emit()
        else:  # pragma: no cover
            pass


@attr.s(auto_attribs=True)
class Window:
    """A manager for a simple window with increment and decrement buttons to change a
    counter which is displayed via a widget in the center.
    """

    widget: QSignaledWidget
    increment: QtWidgets.QPushButton
    decrement: QtWidgets.QPushButton
    label: QtWidgets.QLabel
    layout: QtWidgets.QHBoxLayout
    count: int = 0

    @classmethod
    def build(
        cls,
        title: str = "QTrio Emissions Example",
        parent: typing.Optional[QtWidgets.QWidget] = None,
    ) -> "Window":
        """Build and lay out the widgets that make up this window."""

        self = cls(
            widget=QSignaledWidget(parent),
            layout=QtWidgets.QHBoxLayout(),
            increment=QtWidgets.QPushButton(),
            decrement=QtWidgets.QPushButton(),
            label=QtWidgets.QLabel(),
        )

        self.widget.setWindowTitle(title)
        self.widget.setLayout(self.layout)

        self.increment.setText("+")
        self.decrement.setText("-")

        self.label.setText(str(self.count))

        self.layout.addWidget(self.decrement)
        self.layout.addWidget(self.label)
        self.layout.addWidget(self.increment)

        return self

    def increment_count(self) -> None:
        """Increment the counter and update the label."""

        self.count += 1
        self.label.setText(str(self.count))

    def decrement_count(self) -> None:
        """Decrement the counter and update the label."""

        self.count -= 1
        self.label.setText(str(self.count))

    def show(self) -> None:
        """Show the primary widget for this window."""

        self.widget.show()


async def main(window: typing.Optional[Window] = None) -> None:
    """Show the example window and iterate over the relevant signal emissions to respond
    to user interactions with the GUI.
    """
    if window is None:  # pragma: no cover
        window = Window.build()

    signals = [
        window.decrement.clicked,
        window.increment.clicked,
        window.widget.closed,
    ]

    async with qtrio.enter_emissions_channel(signals=signals) as emissions:
        window.show()

        async for emission in emissions.channel:
            if emission.is_from(window.decrement.clicked):
                window.decrement_count()
            elif emission.is_from(window.increment.clicked):
                window.increment_count()
            elif emission.is_from(window.widget.closed):
                break
            else:  # pragma: no cover
                raise qtrio.QTrioException(f"Unexpected emission: {emission}")



import typing

from qtpy import QtCore
from qtpy import QtWidgets


def create_input() -> QtWidgets.QInputDialog:
    dialog = QtWidgets.QInputDialog()
    dialog.setWindowTitle("Hello")
    dialog.setLabelText("Enter your name:")

    return dialog


def create_output() -> QtWidgets.QMessageBox:
    return QtWidgets.QMessageBox(
        QtWidgets.QMessageBox.Icon.Question,
        "Hello",
        "",
        QtWidgets.QMessageBox.Ok,
    )


class Main:
    def __init__(
        self,
        input_dialog: typing.Optional[QtWidgets.QInputDialog] = None,
        output_dialog: typing.Optional[QtWidgets.QMessageBox] = None,
    ):
        if input_dialog is None:  # pragma: no cover
            input_dialog = create_input()

        if output_dialog is None:  # pragma: no cover
            output_dialog = create_output()

        self.input_dialog = input_dialog
        self.output_dialog = output_dialog

    def setup(self) -> None:
        self.input_dialog.accepted.connect(self.input_accepted)
        self.input_dialog.rejected.connect(self.input_rejected)

        self.input_dialog.show()

    def input_accepted(self) -> None:
        name = self.input_dialog.textValue()

        self.output_dialog.setText(f"Hi {name}, welcome to the team!")

        self.output_dialog.finished.connect(self.output_finished)
        self.output_dialog.show()

    def input_rejected(self) -> None:
        QtCore.QCoreApplication.instance().quit()

    def output_finished(self) -> None:
        QtCore.QCoreApplication.instance().quit()


def main() -> None:  # pragma: no cover
    application = QtWidgets.QApplication([])
    application.setQuitOnLastWindowClosed(False)
    main_object = Main()
    main_object.setup()
    application.exec_()


if __name__ == "__main__":  # pragma: no cover
    main()



import trio

import qtrio.examples.readme.qtrio


async def test_main():
    input_dialog = qtrio.examples.readme.qtrio.create_input()
    output_dialog = qtrio.examples.readme.qtrio.create_output()

    text_to_enter = "everyone"

    async def user(task_status):
        async with qtrio._core.wait_signal_context(input_dialog.shown):
            task_status.started()

        assert input_dialog.line_edit is not None
        input_dialog.line_edit.setText(text_to_enter)

        assert input_dialog.accept_button is not None
        async with qtrio._core.wait_signal_context(output_dialog.shown):
            input_dialog.accept_button.click()

        assert output_dialog.dialog is not None
        output_text = output_dialog.dialog.text()

        assert output_dialog.accept_button is not None
        output_dialog.accept_button.click()

        assert text_to_enter in output_text

    async with trio.open_nursery() as nursery:
        await nursery.start(user)

        await qtrio.examples.readme.qtrio.main(
            input_dialog=input_dialog,
            output_dialog=output_dialog,
        )


async def test_main_cancelled():
    input_dialog = qtrio.examples.readme.qtrio.create_input()
    output_dialog = qtrio.examples.readme.qtrio.create_output()

    text_to_enter = "everyone"

    output_shown = False

    def output_was_shown():
        nonlocal output_shown
        output_shown = True  # pragma: no cover

    output_dialog.shown.connect(output_was_shown)

    async def user(task_status):
        async with qtrio._core.wait_signal_context(input_dialog.shown):
            task_status.started()

        assert input_dialog.line_edit is not None
        input_dialog.line_edit.setText(text_to_enter)

        assert input_dialog.reject_button is not None
        input_dialog.reject_button.click()

    async with trio.open_nursery() as nursery:
        await nursery.start(user)

        await qtrio.examples.readme.qtrio.main(
            input_dialog=input_dialog,
            output_dialog=output_dialog,
        )

    assert not output_shown



"""This module provides general Python related utilities that are not Qt nor Trio
specific.
"""
import typing


# TODO: could support some more things
#       https://docs.python.org/3/library/stdtypes.html#definition.__qualname__
def identifier_path(it: typing.Union[typing.Type[object], typing.Callable]) -> str:
    """Generate an identifier based on an object's module and qualified name.  This can
    be useful such as for adding attributes to existing objects while minimizing odds
    of collisions and maximizing traceability of the related party.

    Args:
        it: The object to generate the identifer from.

    Returns:
        The generated identifier string.
    """
    return "__" + "_".join(it.__module__.split(".") + [it.__qualname__])



import math
import os
import pathlib
import sys

from qtpy import QtCore
from qtpy import QtWidgets
import pytest
import trio


import qtrio
import qtrio._core
import qtrio.dialogs
import qtrio._qt


@pytest.fixture(
    name="builder",
    params=[
        qtrio.dialogs.create_integer_dialog,
        qtrio.dialogs.create_text_input_dialog,
        qtrio.dialogs.create_file_save_dialog,
        qtrio.dialogs.create_file_open_dialog,
        qtrio.dialogs.create_message_box,
        qtrio.dialogs.create_progress_dialog,
    ],
)
def builder_fixture(request):
    yield request.param


async def test_get_integer_gets_value(qtbot):
    dialog = qtrio.dialogs.create_integer_dialog()

    async def user(task_status):
        async with qtrio._core.wait_signal_context(dialog.shown):
            task_status.started()

        qtbot.keyClicks(dialog.edit_widget, str(test_value))
        qtbot.mouseClick(dialog.accept_button, QtCore.Qt.LeftButton)

    test_value = 928

    async with trio.open_nursery() as nursery:
        await nursery.start(user)
        with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
            integer = await dialog.wait()

    assert integer == test_value


async def test_get_integer_raises_cancel_when_canceled(qtbot):
    dialog = qtrio.dialogs.create_integer_dialog()

    async def user(task_status):
        async with qtrio._core.wait_signal_context(dialog.shown):
            task_status.started()

        qtbot.keyClicks(dialog.edit_widget, "abc")
        qtbot.mouseClick(dialog.reject_button, QtCore.Qt.LeftButton)

    async with trio.open_nursery() as nursery:
        await nursery.start(user)
        with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
            with pytest.raises(qtrio.UserCancelledError):
                await dialog.wait()


async def test_get_integer_raises_for_invalid_input(qtbot):
    dialog = qtrio.dialogs.create_integer_dialog()

    async def user(task_status):
        async with qtrio._core.wait_signal_context(dialog.shown):
            task_status.started()

        qtbot.keyClicks(dialog.edit_widget, "abc")
        qtbot.mouseClick(dialog.accept_button, QtCore.Qt.LeftButton)

    async with trio.open_nursery() as nursery:
        await nursery.start(user)
        with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
            with pytest.raises(qtrio.InvalidInputError):
                await dialog.wait()


def test_unused_dialog_teardown_ok(builder):
    dialog = builder()
    dialog.teardown()


async def test_file_save(qtbot, tmp_path):
    path_to_select = trio.Path(tmp_path) / "something.new"

    dialog = qtrio.dialogs.create_file_save_dialog(
        default_directory=path_to_select.parent,
        default_file=path_to_select,
    )

    async def user(task_status):
        async with qtrio._core.wait_signal_context(dialog.shown):
            task_status.started()

        # allow cancellation to occur even if the signal was received before the
        # cancellation was requested.
        await trio.sleep(0)

        assert dialog.dialog is not None

        dialog.dialog.accept()

    async with trio.open_nursery() as nursery:
        await nursery.start(user)
        with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
            selected_path = await dialog.wait()

    assert selected_path == path_to_select


async def test_file_save_no_defaults(qtbot, tmp_path):
    path_to_select = trio.Path(tmp_path) / "another.thing"

    dialog = qtrio.dialogs.create_file_save_dialog()

    async def user(task_status):
        async with qtrio._core.wait_signal_context(dialog.shown):
            task_status.started()

        # allow cancellation to occur even if the signal was received before the
        # cancellation was requested.
        await trio.sleep(0)

        assert dialog.dialog is not None

        dialog.dialog.setDirectory(os.fspath(path_to_select.parent))
        [text_edit] = dialog.dialog.findChildren(QtWidgets.QLineEdit)
        text_edit.setText(path_to_select.name)
        dialog.dialog.accept()

    async with trio.open_nursery() as nursery:
        await nursery.start(user)
        with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
            selected_path = await dialog.wait()

    assert selected_path == path_to_select


async def test_file_save_cancelled(qtbot, tmp_path):
    dialog = qtrio.dialogs.create_file_save_dialog()

    async def user(task_status):
        async with qtrio._core.wait_signal_context(dialog.shown):
            task_status.started()

        # allow cancellation to occur even if the signal was received before the
        # cancellation was requested.
        await trio.sleep(0)

        assert dialog.dialog is not None

        dialog.dialog.reject()

    async with trio.open_nursery() as nursery:
        await nursery.start(user)
        with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
            with pytest.raises(qtrio.UserCancelledError):
                await dialog.wait()


async def test_file_open_set_path(tmp_path: pathlib.Path) -> None:
    file_path = tmp_path.joinpath("some_file")
    file_path.touch()

    dialog = qtrio.dialogs.create_file_open_dialog()

    trio_file_path = trio.Path(file_path)

    async def user():
        await emissions.channel.receive()

        await dialog.set_path(path=trio_file_path)

        assert dialog.accept_button is not None
        dialog.accept_button.click()

    async with qtrio.enter_emissions_channel(signals=[dialog.shown]) as emissions:
        async with trio.open_nursery() as nursery:
            nursery.start_soon(user)

            selected_path = await dialog.wait()

    assert selected_path == trio_file_path


async def test_file_save_raises_for_path_selection_when_not_active(qtbot):
    dialog = qtrio.dialogs.create_file_save_dialog()

    with pytest.raises(qtrio.DialogNotActiveError):
        await dialog.set_path(path=trio.Path())


async def test_information_message_box(qtbot):
    text = "Consider yourself informed."
    queried_text = None

    dialog = qtrio.dialogs.create_message_box(
        title="Information",
        text=text,
        icon=QtWidgets.QMessageBox.Information,
    )

    async def user(task_status):
        nonlocal queried_text

        async with qtrio._core.wait_signal_context(dialog.shown):
            task_status.started()

        assert dialog.dialog is not None

        queried_text = dialog.dialog.text()
        dialog.dialog.accept()

    async with trio.open_nursery() as nursery:
        await nursery.start(user)
        with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
            await dialog.wait()

    assert queried_text == text


async def test_information_message_box_cancel(qtbot):
    dialog = qtrio.dialogs.create_message_box(
        title="",
        text="",
        icon=QtWidgets.QMessageBox.Information,
        buttons=QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel,
    )

    async def user(task_status):
        async with qtrio._core.wait_signal_context(dialog.shown):
            task_status.started()

        assert dialog.dialog is not None

        dialog.dialog.reject()

    async with trio.open_nursery() as nursery:
        await nursery.start(user)
        with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
            with pytest.raises(qtrio.UserCancelledError):
                await dialog.wait()


async def test_text_input_dialog(qtbot):
    dialog = qtrio.dialogs.create_text_input_dialog()

    entered_text = "etcetera"

    async def user(task_status):
        async with qtrio._core.wait_signal_context(dialog.shown):
            task_status.started()

        qtbot.keyClicks(dialog.line_edit, entered_text)

        assert dialog.dialog is not None

        dialog.dialog.accept()

    async with trio.open_nursery() as nursery:
        await nursery.start(user)
        with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
            returned_text = await dialog.wait()

    assert returned_text == entered_text


def test_text_input_dialog_with_title():
    title_string = "abc123"
    dialog = qtrio.dialogs.create_text_input_dialog(title=title_string)
    with qtrio.dialogs._manage(dialog=dialog):
        assert dialog.dialog is not None

        assert dialog.dialog.windowTitle() == title_string


def test_text_input_dialog_with_label():
    label_string = "lmno789"
    dialog = qtrio.dialogs.create_text_input_dialog(label=label_string)
    with qtrio.dialogs._manage(dialog=dialog):
        assert dialog.dialog is not None

        [label] = dialog.dialog.findChildren(QtWidgets.QLabel)
        assert label.text() == label_string


async def test_text_input_dialog_cancel(qtbot):
    dialog = qtrio.dialogs.create_text_input_dialog()

    async def user(task_status):
        async with qtrio._core.wait_signal_context(dialog.shown):
            task_status.started()

        assert dialog.dialog is not None

        dialog.dialog.reject()

    async with trio.open_nursery() as nursery:
        await nursery.start(user)
        with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
            with pytest.raises(qtrio.UserCancelledError):
                await dialog.wait()


async def test_progress_dialog_dot_dot_dot(qtbot):
    dialog = qtrio.dialogs.create_progress_dialog()

    with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
        async with dialog.manage():
            pass


async def test_progress_dialog_cancel_raises(qtbot):
    dialog = qtrio.dialogs.create_progress_dialog(cancel_button_text="cancel here")

    with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
        with pytest.raises(qtrio.UserCancelledError):
            async with dialog.manage():
                assert dialog.dialog is not None
                dialog.dialog.cancel()


async def test_progress_dialog_cancel_cancels_context(qtbot):
    dialog = qtrio.dialogs.create_progress_dialog(cancel_button_text="cancel here")

    cancelled = False

    with pytest.raises(qtrio.UserCancelledError):
        with qtrio._qt.connection(signal=dialog.shown, slot=qtbot.addWidget):
            async with dialog.manage():
                try:
                    assert dialog.cancel_button is not None
                    dialog.cancel_button.click()
                    await trio.sleep(math.inf)
                except trio.Cancelled:
                    cancelled = True
                    raise

    assert cancelled


def test_dialog_button_box_buttons_by_role_no_buttons(qtbot):
    dialog = QtWidgets.QDialog()
    assert qtrio.dialogs._dialog_button_box_buttons_by_role(dialog=dialog) == {}



"""Thanks to njsmith for developing Trio guest mode and sharing the first example
integration with Qt.
"""
import contextlib
import math
import os
import time
import typing

import attr
import httpcore._async.http11
import httpx
import hyperlink
import qtrio
import trio

import qtrio.dialogs


# Default is 4096
httpcore._async.http11.AsyncHTTP11Connection.READ_NUM_BYTES = 100_000


def create_title(specific: str) -> str:
    return f"QTrio Download Example - {specific}"


@attr.s(auto_attribs=True, frozen=True)
class Progress:
    downloaded: int
    total: typing.Optional[int]
    first: bool


async def main(
    url: typing.Optional[typing.Union[str, hyperlink.URL]],
    destination: typing.Optional[typing.Union[str, os.PathLike]],
    fps: int = 60,
    text_input_dialog: typing.Optional[qtrio.dialogs.TextInputDialog] = None,
    file_dialog: typing.Optional[qtrio.dialogs.FileDialog] = None,
    progress_dialog: typing.Optional[qtrio.dialogs.ProgressDialog] = None,
    message_box: typing.Optional[qtrio.dialogs.MessageBox] = None,
    clock: typing.Callable[[], float] = time.monotonic,
    http_application: typing.Optional[typing.Callable[..., typing.Any]] = None,
) -> None:
    print("file dialog right inside", id(file_dialog), file_dialog, flush=True)
    converted_url: hyperlink.URL
    converted_destination: trio.Path

    with contextlib.suppress(qtrio.UserCancelledError):
        if url is None:
            if text_input_dialog is None:  # pragma: no cover
                text_input_dialog = qtrio.dialogs.create_text_input_dialog()

            text_input_dialog.title = create_title("Enter URL")
            text_input_dialog.label = "URL to download:"

            url = await text_input_dialog.wait()

        if isinstance(url, str):
            converted_url = hyperlink.URL.from_text(url)
        else:
            converted_url = url

        if destination is None:
            if file_dialog is None:  # pragma: no cover
                file_dialog = qtrio.dialogs.create_file_save_dialog()

            file_dialog.default_file = trio.Path(converted_url.path[-1])

            destination = await file_dialog.wait()

        converted_destination = trio.Path(destination)

        await get_dialog(
            url=converted_url,
            destination=converted_destination,
            fps=fps,
            progress_dialog=progress_dialog,
            message_box=message_box,
            clock=clock,
            http_application=http_application,
        )


async def get_dialog(
    url: hyperlink.URL,
    destination: trio.Path,
    fps: float,
    progress_dialog: typing.Optional[qtrio.dialogs.ProgressDialog] = None,
    message_box: typing.Optional[qtrio.dialogs.MessageBox] = None,
    clock: typing.Callable[[], float] = time.monotonic,
    http_application: typing.Optional[typing.Callable[..., typing.Any]] = None,
) -> None:
    if progress_dialog is None:  # pragma: no cover
        progress_dialog = qtrio.dialogs.create_progress_dialog()

    progress_dialog.title = create_title("Fetching")
    progress_dialog.text = f"Fetching {url}..."

    async with progress_dialog.manage():
        if progress_dialog.dialog is None:  # pragma: no cover
            raise qtrio.InternalError("Dialog not assigned while it is being managed.")

        # Always show the dialog
        progress_dialog.dialog.setMinimumDuration(0)

        start = clock()

        async for progress in get(
            url=url,
            destination=destination,
            update_period=1 / fps,
            http_application=http_application,
        ):
            if progress.first:
                if progress.total is None:
                    maximum = 0
                else:
                    maximum = progress.total

                progress_dialog.dialog.setMaximum(maximum)
                progress_dialog.dialog.setValue(0)

            if progress.total is not None:
                progress_dialog.dialog.setValue(progress.downloaded)

        end = clock()

    duration = end - start
    if duration == 0:
        # define this seems to happen when testing on Windows with an x86 Python
        if progress.downloaded > 0:
            bytes_per_second = math.inf
        else:  # pragma: no cover
            bytes_per_second = 0
    else:
        bytes_per_second = progress.downloaded / duration

    summary = "\n\n".join(
        [
            url.asText(),
            os.fspath(destination),
            f"Downloaded {progress.downloaded} bytes in {duration:.2f} seconds",
            f"{bytes_per_second:.2f} bytes/second",
        ]
    )

    if message_box is None:  # pragma: no cover
        message_box = qtrio.dialogs.create_message_box()

    message_box.title = create_title("Download Summary")
    message_box.text = summary

    await message_box.wait()


async def get(
    url: hyperlink.URL,
    destination: trio.Path,
    update_period: float,
    clock: typing.Callable[[], float] = time.monotonic,
    http_application: typing.Optional[typing.Callable[..., typing.Any]] = None,
) -> typing.AsyncIterable[Progress]:
    async with httpx.AsyncClient(app=http_application) as client:
        async with client.stream("GET", url.asText()) as response:
            raw_content_length = response.headers.get("content-length")
            if raw_content_length is None:
                content_length = None
            else:
                content_length = int(raw_content_length)

            progress = Progress(
                downloaded=0,
                total=content_length,
                first=True,
            )

            yield progress
            last_update = clock()

            progress = attr.evolve(progress, first=False)

            downloaded = 0

            async with (await destination.open("wb")) as file:
                async for chunk in response.aiter_raw():
                    downloaded += len(chunk)
                    await file.write(chunk)  # type: ignore[attr-defined]

                    if clock() - last_update > update_period:
                        progress = attr.evolve(progress, downloaded=downloaded)
                        yield progress
                        last_update = clock()

            if progress.downloaded != downloaded:
                progress = attr.evolve(progress, downloaded=downloaded)
                yield progress



import click


@click.group()
def cli() -> None:
    """QTrio - a library bringing Qt GUIs together with async and await via Trio"""

    pass


@cli.group()
def examples() -> None:
    """Run code examples."""

    pass


@examples.command()
def emissions() -> None:  # pragma: no cover
    """A simple demonstration of iterating over signal emissions."""

    import qtrio
    import qtrio.examples.emissions

    qtrio.run(qtrio.examples.emissions.main)


@examples.command()
@click.option("--url", help="The URL to download.")
@click.option(
    "--destination",
    help="The file path to save to.",
    type=click.Path(dir_okay=False, writable=True, resolve_path=True, allow_dash=True),
)
@click.option(
    "--fps",
    default=60,
    help="Frames per second for progress updates.",
    type=click.IntRange(min=1),
)
def download(url: str, destination: str, fps: int) -> None:  # pragma: no cover
    import qtrio
    import qtrio.examples.download

    qtrio.run(qtrio.examples.download.main, url, destination, fps)



"""This module provides general Qt related utilities that are not Trio specific."""

import contextlib
import typing

from qtpy import QtCore

import qtrio._python
import qtrio._util


class Signal:
    """This is a (nearly) drop-in replacement for :class:`QtCore.Signal`.  The useful
    difference is that it does not require inheriting from :class:`QtCore.QObject`.  The
    not-quite part is that it will be a bit more complicated to change thread affinity
    of the relevant :class:`QtCore.QObject`.  If you need this, maybe just inherit.

    This signal gets around the normally required inheritance by creating
    :class:`QtCore.QObject` instances behind the scenes to host the real signals.  Just
    as :class:`QtCore.Signal` uses the Python descriptor protocol to intercept the
    attribute access, so does this so it can 'redirect' to the signal on the other
    object.
    """

    _attribute_name: typing.ClassVar[str] = ""

    def __init__(self, *args: object, **kwargs: object) -> None:
        class _SignalQObject(QtCore.QObject):
            signal = QtCore.Signal(*args, **kwargs)

        self.object_cls = _SignalQObject

    def __get__(self, instance: object, owner: object) -> qtrio._util.SignalInstance:
        if instance is None:
            return self

        o = self.object(instance=instance)

        return o.signal

    def object(self, instance: object) -> QtCore.QObject:
        """Get the :class:`QtCore.QObject` that hosts the real signal.  This can be
        called such as ``type(instance).signal_name.object(instance)``.  Yes this is
        non-obvious but you have to do something special to get around the
        :ref:`descriptor protocol <python:descriptors>` so you can get at this method
        instead of just having the underlying :class:`QtCore.SignalInstance`.

        Arguments:
            instance: The object on which this descriptor instance is hosted.

        Returns:
            The signal-hosting :class:`QtCore.QObject`.
        """
        d = getattr(instance, self._attribute_name, None)

        if d is None:
            d = {}
            setattr(instance, self._attribute_name, d)

        o = d.get(self.object_cls)
        if o is None:
            o = self.object_cls()
            d[self.object_cls] = o

        return o


Signal._attribute_name = qtrio._python.identifier_path(Signal)


@contextlib.contextmanager
def connection(
    signal: qtrio._util.SignalInstance, slot: typing.Callable[..., object]
) -> typing.Generator[
    typing.Union[QtCore.QMetaObject.Connection, typing.Callable[..., object]],
    None,
    None,
]:
    """Connect a signal and slot for the duration of the context manager.

    Args:
        signal: The signal to connect.
        slot: The callable to connect the signal to.
    """

    # if you get segfault or sigsegv here, especially from pyside2<5.15.2, make
    # sure the slot isn't on a non-hashable (frozen will make it hashable) attrs
    # class.  https://bugreports.qt.io/browse/PYSIDE-1422
    this_connection = signal.connect(slot)

    import qtpy

    if qtpy.PYSIDE2:
        # PySide2 presently returns a bool rather than a QMetaObject.Connection
        # https://bugreports.qt.io/browse/PYSIDE-1334
        this_connection = slot

    try:
        yield this_connection
    finally:
        expected_exception: typing.Type[Exception]

        if qtpy.PYSIDE2:
            expected_exception = RuntimeError
        else:
            expected_exception = TypeError

        try:
            # can we precheck and avoid the exception?
            signal.disconnect(this_connection)
        except expected_exception:
            pass



"""The module holding the core features of QTrio.

Attributes:
    _reenter_event_type: The event type enumerator for our reenter events.
"""
import contextlib
import functools
import math
import sys
import traceback
import typing
import typing_extensions

import async_generator
import attr
import outcome
from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets
import trio
import trio.abc

import qtrio
import qtrio._qt
import qtrio._util

_reenter_event_type: typing.Optional[QtCore.QEvent.Type] = None


def registered_event_type() -> typing.Optional[QtCore.QEvent.Type]:
    """Get the registered event type.

    Returns:
        The type registered with Qt for the reenter event.  :obj:`None` if no event type
        has been registered yet.
    """
    return _reenter_event_type


def register_event_type() -> None:
    """Register a Qt event type for use by Trio to reenter into the Qt event loop.

    Raises:
        qtrio.EventTypeAlreadyRegisteredError: if an event type has already been
            registered.
        qtrio.EventTypeRegistrationFailedError:  if a type was not able to be
            registered.
    """
    global _reenter_event_type

    if _reenter_event_type is not None:
        raise qtrio.EventTypeAlreadyRegisteredError()

    event_hint = QtCore.QEvent.registerEventType()

    if event_hint == -1:
        raise qtrio.EventTypeRegistrationFailedError()

    # assign to the global
    _reenter_event_type = QtCore.QEvent.Type(event_hint)


def register_requested_event_type(
    requested_value: typing.Union[int, QtCore.QEvent.Type]
) -> None:
    """Register the requested Qt event type for use by Trio to reenter into the Qt event
    loop.

    Arguments:
        requested_value: The value to ask Qt to use for the event type being registered.

    Raises:
        qtrio.EventTypeAlreadyRegisteredError: if an event type has already been
            registered.
        qtrio.EventTypeRegistrationFailedError: if a type was not able to be registered.
        qtrio.RequestedEventTypeUnavailableError: if the type returned by Qt does not
            match the requested type.
    """
    global _reenter_event_type

    if _reenter_event_type is not None:
        raise qtrio.EventTypeAlreadyRegisteredError()

    event_hint = QtCore.QEvent.registerEventType(requested_value)

    if event_hint == -1:
        raise qtrio.EventTypeRegistrationFailedError()
    elif event_hint != requested_value:
        raise qtrio.RequestedEventTypeUnavailableError(
            requested_type=requested_value, returned_type=event_hint
        )

    # assign to the global
    _reenter_event_type = QtCore.QEvent.Type(event_hint)


class ReenterEvent(QtCore.QEvent):
    """A proper ``ReenterEvent`` for reentering into the Qt host loop."""

    def __init__(self, fn: typing.Callable[[], object]):
        super().__init__(_reenter_event_type)
        self.fn = fn


class Reenter(QtCore.QObject):
    """A ``QtCore.QObject`` for handling reenter events."""

    def event(self, event: QtCore.QEvent) -> bool:
        """Qt calls this when the object receives an event."""

        reenter_event = typing.cast(Reenter, event)
        reenter_event.fn()
        return False


async def wait_signal(signal: qtrio._util.SignalInstance) -> typing.Tuple[object, ...]:
    """Block for the next emission of ``signal`` and return the emitted arguments.

    Warning:
        In many cases this can result in a race condition since you are unable to
        first connect the signal and then wait for it.

    Args:
        signal: The signal instance to wait for emission of.

    Returns:
        A tuple containing the values emitted by the signal.
    """
    event = trio.Event()
    result: typing.Tuple[object, ...] = ()

    def slot(*args: object) -> None:
        """Receive and store the emitted arguments and set the event so we can continue.

        Args:
            args: The arguments emitted from the signal.
        """
        nonlocal result
        result = args
        event.set()

    with qtrio._qt.connection(signal, slot):
        await event.wait()

    return result


@attr.s(auto_attribs=True, frozen=True, slots=True, eq=False)
class Emission:
    """Stores the emission of a signal including the emitted arguments.  Can be
    compared against a signal instance to check the source.  Do not construct this class
    directly.  Instead, instances will be received through a channel created by
    :func:`qtrio.enter_emissions_channel`.

    Note:
        Each time you access a signal such as ``a_qobject.some_signal`` you get a
        different signal instance object so the ``signal`` attribute generally will not
        be the same object.  A signal instance is a ``QtCore.SignalInstance`` in
        PySide2 or ``QtCore.pyqtBoundSignal`` in PyQt5.
    """

    signal: qtrio._util.SignalInstance
    """An instance of the original signal."""
    args: typing.Tuple[object, ...]
    """A tuple of the arguments emitted by the signal."""

    def is_from(self, signal: qtrio._util.SignalInstance) -> bool:
        """Check if this emission came from ``signal``.

        Args:
            signal: The signal instance to check for being the source.

        Returns:
            Whether the passed signal was the source of this emission.
        """

        # bool() to accomodate SignalInstance being typed Any right now...
        return bool(self.signal == signal)

    def __eq__(self, other: object) -> bool:
        if type(other) != type(self):
            return False

        # TODO: workaround for https://github.com/python/mypy/issues/4445
        if not isinstance(other, type(self)):  # pragma: no cover
            return False

        return self.is_from(signal=other.signal) and self.args == other.args


@attr.s(auto_attribs=True)
class Emissions:
    """Hold elements useful for the application to work with emissions from signals.
    Do not construct this class directly.  Instead, use
    :func:`qtrio.enter_emissions_channel`.
    """

    channel: trio.MemoryReceiveChannel
    """A memory receive channel to be fed by signal emissions."""
    send_channel: trio.MemorySendChannel
    """A memory send channel collecting signal emissions."""

    async def aclose(self) -> None:
        """Asynchronously close the send channel when signal emissions are no longer of
        interest.
        """
        await self.send_channel.aclose()


@async_generator.asynccontextmanager
async def open_emissions_channel(
    signals: typing.Collection[qtrio._util.SignalInstance],
    max_buffer_size: typing.Union[int, float] = math.inf,
) -> typing.AsyncGenerator[Emissions, None]:
    """Create a memory channel fed by the emissions of the signals.  Each signal
    emission will be converted to a :class:`qtrio.Emission` object.  On exit the send
    channel is closed.  Management of the receive channel is left to the caller.

    Note:
        Use this only if you need to process emissions *after* exiting the context
        manager.  Otherwise use :func:`qtrio.enter_emissions_channel`.

    Args:
        signals: A collection of signals which will be monitored for emissions.
        max_buffer_size: When the number of unhandled emissions in the channel reaches
            this limit then additional emissions will be silently thrown out the window.

    Returns:
        The emissions manager with the signals connected to it.
    """

    # Infinite buffer because I don't think there's any use in storing the emission
    # info in a `slot()` stack frame rather than in the memory channel.  Perhaps in the
    # future we can implement a limit beyond which events are thrown away to avoid
    # infinite queueing.  Maybe trio.MemorySendChannel.send_nowait() instead.
    send_channel, receive_channel = trio.open_memory_channel[Emission](
        max_buffer_size=max_buffer_size
    )

    async with send_channel:
        with contextlib.ExitStack() as stack:
            for signal in signals:

                def slot(
                    *args: object, internal_signal: qtrio._util.SignalInstance = signal
                ) -> None:
                    try:
                        send_channel.send_nowait(
                            Emission(signal=internal_signal, args=args)
                        )
                    except trio.WouldBlock:
                        # TODO: log this or... ?
                        pass

                stack.enter_context(qtrio._qt.connection(signal, slot))

            yield Emissions(channel=receive_channel, send_channel=send_channel)


@async_generator.asynccontextmanager
async def enter_emissions_channel(
    signals: typing.Collection[qtrio._util.SignalInstance],
    max_buffer_size: typing.Union[int, float] = math.inf,
) -> typing.AsyncGenerator[Emissions, None]:
    """Create a memory channel fed by the emissions of the signals and enter both the
    send and receive channels' context managers.

    Args:
        signals: A collection of signals which will be monitored for emissions.
        max_buffer_size: When the number of unhandled emissions in the channel reaches
            this limit then additional emissions will be silently thrown out the window.

    Returns:
        The emissions manager.
    """
    async with open_emissions_channel(
        signals=signals, max_buffer_size=max_buffer_size
    ) as emissions:
        async with emissions.channel:
            async with emissions.send_channel:
                yield emissions


class StarterProtocol(typing_extensions.Protocol):
    def start(self, *args: object) -> None:
        ...


@attr.s(auto_attribs=True, frozen=True)
class DirectStarter:
    slot: typing.Callable[..., typing.Awaitable[object]]
    nursery: trio.Nursery

    def start(self, *args: object) -> None:
        self.nursery.start_soon(self.slot, *args)


@attr.s(auto_attribs=True, frozen=True)
class WrappedStarter:
    slot: typing.Callable[..., typing.Awaitable[object]]
    wrapper: typing.Callable[
        [typing.Callable[..., typing.Awaitable[object]]], typing.Awaitable[object]
    ]
    nursery: trio.Nursery

    def start(self, *args: object) -> None:
        self.nursery.start_soon(self.wrapper, self.slot, *args)


@attr.s(auto_attribs=True)
class EmissionsNursery:
    """Holds the nursery, exit stack, and wrapper needed to support connecting signals
    to both async and sync slots in the nursery.
    """

    nursery: trio.Nursery
    """The Trio nursery that will handle execution of the slots."""
    exit_stack: contextlib.ExitStack
    """The exit stack that will manage the connections so they get disconnected."""
    wrapper: typing.Optional[
        typing.Callable[
            [typing.Callable[..., typing.Awaitable[object]]],
            typing.Awaitable[object],
        ]
    ] = None
    """The wrapper for handling the slots.  This could, for example, handle exceptions
    and present a dialog to avoid cancelling the entire nursery.
    """

    def connect(
        self,
        signal: qtrio._util.SignalInstance,
        slot: typing.Callable[..., typing.Awaitable[object]],
    ) -> None:
        """Connect an async signal to this emissions nursery so when called the slot
        will be run in the nursery.
        """
        starter: StarterProtocol

        if self.wrapper is None:
            starter = DirectStarter(slot=slot, nursery=self.nursery)
        else:
            starter = WrappedStarter(
                slot=slot, wrapper=self.wrapper, nursery=self.nursery
            )

        self.exit_stack.enter_context(qtrio._qt.connection(signal, starter.start))

    def connect_sync(
        self, signal: qtrio._util.SignalInstance, slot: typing.Callable[..., object]
    ) -> None:
        """Connect to a sync slot to this emissions nursery so when called the slot will
        be run in the nursery.
        """

        async def async_slot(*args: object) -> None:
            slot(*args)

        self.connect(signal=signal, slot=async_slot)


@async_generator.asynccontextmanager
async def open_emissions_nursery(
    until: typing.Optional[qtrio._util.SignalInstance] = None,
    wrapper: typing.Optional[typing.Callable[..., typing.Awaitable[object]]] = None,
) -> typing.AsyncGenerator[EmissionsNursery, None]:
    """Open a nursery for handling callbacks triggered by signal emissions.  This allows
    a 'normal' Qt callback structure while still executing the callbacks within a Trio
    nursery such that errors have a place to go.  Both async and sync callbacks can be
    connected.  Sync callbacks will be wrapped in an async call to allow execution in
    the nursery.

    Arguments:
        until: Keep the nursery open until this signal is emitted.
        wrapper: A wrapper for the callbacks such as to process exceptions.

    Returns:
        The emissions manager.
    """
    async with trio.open_nursery() as nursery:
        with contextlib.ExitStack() as exit_stack:
            emissions_nursery = EmissionsNursery(
                nursery=nursery,
                exit_stack=exit_stack,
                wrapper=wrapper,
            )

            if until is not None:
                async with wait_signal_context(until):
                    yield emissions_nursery
            else:
                yield emissions_nursery


@async_generator.asynccontextmanager
async def wait_signal_context(
    signal: qtrio._util.SignalInstance,
) -> typing.AsyncGenerator[None, None]:
    """Connect a signal during the context and wait for it on exit.  Presently no
    mechanism is provided for retrieving the emitted arguments.

    Args:
        signal: The signal to connect to and wait for.
    """
    event = trio.Event()

    def slot(*args: object, **kwargs: object) -> None:
        event.set()

    with qtrio._qt.connection(signal=signal, slot=slot):
        yield
        await event.wait()


@attr.s(auto_attribs=True, frozen=True, slots=True)
class Outcomes:
    """This class holds an :class:`outcome.Outcome` from each of the Trio and the Qt
    application execution.  Do not construct instances directly.  Instead, an instance
    will be returned from :func:`qtrio.run` or available on instances of
    :attr:`qtrio.Runner.outcomes`.
    """

    qt: typing.Optional[outcome.Outcome] = None
    """The Qt application :class:`outcome.Outcome`"""
    trio: typing.Optional[outcome.Outcome] = None
    """The Trio async function :class:`outcome.Outcome`"""

    def unwrap(self) -> object:
        """Unwrap either the Trio or Qt outcome.  First, errors are given priority over
        success values.  Second, the Trio outcome gets priority over the Qt outcome.

        Returns:
            Whatever captured value was selected.

        Raises:
            Exception: Whatever captured exception was selected.
            qtrio.NoOutcomesError: if no value or exception has been captured.
        """

        if self.trio is not None:
            # highest priority to the Trio outcome, if it is an error we are done
            result = self.trio.unwrap()

            # since a Trio result is higher priority, we only care if Qt gave an error
            if self.qt is not None:
                self.qt.unwrap()

            # no Qt error so go ahead and return the Trio result
            return result
        elif self.qt is not None:
            # either it is a value that gets returned or an error that gets raised
            return self.qt.unwrap()

        # neither Trio nor Qt outcomes have been set so we have nothing to unwrap()
        raise qtrio.NoOutcomesError()


def run(
    async_fn: typing.Callable[..., typing.Awaitable[object]],
    *args: object,
    done_callback: typing.Optional[typing.Callable[[Outcomes], None]] = None,
    clock: typing.Optional[trio.abc.Clock] = None,
    instruments: typing.Sequence[trio.abc.Instrument] = (),
) -> object:
    """Run a Trio-flavored async function in guest mode on a Qt host application, and
    return the result.

    Args:
        async_fn: The async function to run.
        args: Positional arguments to pass to `async_fn`.
        done_callback: See :class:`qtrio.Runner.done_callback`.
        clock: See :class:`qtrio.Runner.clock`.
        instruments: See :class:`qtrio.Runner.instruments`.

    Returns:
        The object returned by ``async_fn``.
    """
    runner = Runner(
        done_callback=done_callback, clock=clock, instruments=list(instruments)
    )
    runner.run(async_fn, *args)

    return runner.outcomes.unwrap()


def outcome_from_application_return_code(return_code: int) -> outcome.Outcome:
    """Create either an :class:`outcome.Value` in the case of a 0 `return_code` or an
    :class:`outcome.Error` with a :class:`ReturnCodeError` otherwise.

    Args:
        return_code: The return code to be processed.

    Returns:
        The outcome wrapping the passed in return code.
    """

    if return_code == 0:
        return outcome.Value(return_code)

    return outcome.Error(qtrio.ReturnCodeError(return_code))


def maybe_build_application() -> QtGui.QGuiApplication:
    """Create a new Qt application object if one does not already exist.

    Returns:
        The Qt application object.
    """
    maybe_application = QtWidgets.QApplication.instance()
    if maybe_application is None:
        application = QtWidgets.QApplication(sys.argv[1:])
    else:
        application = maybe_application

    application.setQuitOnLastWindowClosed(False)

    return application


@attr.s(auto_attribs=True, slots=True)
class Runner:
    """This class helps run Trio in guest mode on a Qt host application."""

    application: QtGui.QGuiApplication = attr.ib(factory=maybe_build_application)
    """The Qt application object to run as the host.  If not set before calling
    :meth:`run` the application will be created as
    ``QtWidgets.QApplication(sys.argv[1:])`` and ``.setQuitOnLastWindowClosed(False)``
    will be called on it to allow the application to continue throughout the lifetime of
    the async function passed to :meth:`qtrio.Runner.run`.
    """
    quit_application: bool = True
    """When true, the :meth:`done_callback` method will quit the application when the
    async function passed to :meth:`qtrio.Runner.run` has completed.
    """
    clock: typing.Optional[trio.abc.Clock] = None
    """The clock to use for this run.  This is primarily used to speed up tests that
    include timeouts.  The value will be passed on to
    :func:`trio.lowlevel.start_guest_run`.
    """
    instruments: typing.Sequence[trio.abc.Instrument] = ()
    """The instruments to use for this run.  The value will be passed on to
    :func:`trio.lowlevel.start_guest_run`.
    """

    reenter: Reenter = attr.ib(factory=Reenter)
    """The :class:`QtCore.QObject` instance which will receive the events requesting
    execution of the needed Trio and user code in the host's event loop and thread.
    """

    done_callback: typing.Optional[typing.Callable[[Outcomes], None]] = attr.ib(
        default=None
    )
    """The builtin :meth:`done_callback` will be passed to
    :func:`trio.lowlevel.start_guest_run` but will call the callback passed here before
    (maybe) quitting the application.  The :class:`outcome.Outcome` from the completion
    of the async function passed to :meth:`run` will be passed to this callback.
    """

    outcomes: Outcomes = attr.ib(factory=Outcomes, init=False)
    """The outcomes from the Qt and Trio runs."""
    cancel_scope: trio.CancelScope = attr.ib(default=None, init=False)
    """An all encompassing cancellation scope for the Trio execution."""

    _done: bool = attr.ib(default=False, init=False)
    """Just an indicator that the run is done.  Presently used only for a test."""

    def run(
        self,
        async_fn: typing.Callable[..., typing.Awaitable[object]],
        *args: object,
        execute_application: bool = True,
    ) -> Outcomes:
        """Start the guest loop executing ``async_fn``.

        Args:
            async_fn: The async function to be run in the Qt host loop by the Trio
                guest.
            args: Arguments to pass when calling ``async_fn``.
            execute_application: If True, the Qt application will be executed and this
                call will block until it finishes.

        Returns:
            If ``execute_application`` is true, a :class:`qtrio.Outcomes` containing
            outcomes from the Qt application and ``async_fn`` will be returned.
            Otherwise, an empty :class:`qtrio.Outcomes`.
        """
        if _reenter_event_type is None:
            register_event_type()

        trio.lowlevel.start_guest_run(
            self.trio_main,
            async_fn,
            args,
            run_sync_soon_threadsafe=self.run_sync_soon_threadsafe,
            done_callback=self.trio_done,
            clock=self.clock,  # type: ignore[arg-type]
            instruments=self.instruments,
        )

        if execute_application:
            return_code = self.application.exec_()

            self.outcomes = attr.evolve(
                self.outcomes,
                qt=outcome_from_application_return_code(return_code),
            )

        return self.outcomes

    def run_sync_soon_threadsafe(self, fn: typing.Callable[[], object]) -> None:
        """Helper for the Trio guest to execute a sync function in the Qt host
        thread when called from the Trio guest thread.  This call will not block waiting
        for completion of ``fn`` nor will it return the result of calling ``fn``.

        Args:
            fn: A no parameter callable.
        """
        event = ReenterEvent(fn=fn)
        self.application.postEvent(self.reenter, event)

    async def trio_main(
        self,
        async_fn: typing.Callable[..., typing.Awaitable[object]],
        args: typing.Tuple[object, ...],
    ) -> object:
        """Will be run as the main async function by the Trio guest.  It creates a
        cancellation scope to be cancelled when
        :meth:`QtGui.QGuiApplication.lastWindowClosed` is emitted.  Within this scope
        the application's ``async_fn`` will be run and passed ``args``.

        Args:
            async_fn: The application's main async function to be run by Trio in the Qt
                host's thread.
            args: Positional arguments to be passed to ``async_fn``

        Returns:
            The result returned by `async_fn`.
        """
        result: object = None

        with trio.CancelScope() as self.cancel_scope:
            with contextlib.ExitStack() as exit_stack:
                if self.application.quitOnLastWindowClosed():
                    exit_stack.enter_context(
                        qtrio._qt.connection(
                            signal=self.application.lastWindowClosed,
                            slot=self.cancel_scope.cancel,
                        )
                    )

                result = await async_fn(*args)

        return result

    def trio_done(self, run_outcome: outcome.Outcome) -> None:
        """Will be called after the Trio guest run has finished.  This allows collection
        of the :class:`outcome.Outcome` and execution of any application provided done
        callback.  Finally, if :attr:`qtrio.Runner.quit_application` was set when
        creating the instance then the Qt application will be requested to quit.

        Actions such as outputting error information or unwrapping the outcomes need
        to be further considered.

        Arguments:
            run_outcome: The outcome of the Trio guest run.
        """
        self.outcomes = attr.evolve(self.outcomes, trio=run_outcome)

        if self.done_callback is not None:
            self.done_callback(self.outcomes)

        if self.quit_application:
            self.application.quit()

        self._done = True



import typing

import qtrio
from qtpy import QtCore
from qtpy import QtWidgets
import trio

import qtrio.examples.buildingrespect


async def test_main(request, qtbot):
    class SignaledLabel(QtWidgets.QLabel):
        text_changed = QtCore.Signal(str)

        def setText(self, *args, **kwargs):
            super().setText(*args, **kwargs)
            self.text_changed.emit(self.text())

    widget = qtrio.examples.buildingrespect.Widget(label=SignaledLabel())
    qtbot.addWidget(widget.widget)

    message = "test world"
    results: typing.List[str] = []

    async def user():
        for _ in message:
            widget.button.click()

        with trio.move_on_after(1):
            async for emission in emissions.channel:
                [text] = emission.args
                results.append(text)

        widget.button.click()

    async with qtrio.enter_emissions_channel(
        signals=[widget.label.text_changed],
    ) as emissions:
        async with trio.open_nursery() as nursery:
            nursery.start_soon(user)

            await qtrio.examples.buildingrespect.main(
                widget=widget,
                message=message,
            )

    assert results == [
        "test world",
        "",
        "t",
        "te",
        "tes",
        "test",
        "test ",
        "test w",
        "test wo",
        "test wor",
        "test worl",
        "test world",
    ]



import contextlib
import typing

from qtpy import QtWidgets

import qtrio
import qtrio.dialogs


def create_input() -> qtrio.dialogs.TextInputDialog:
    return qtrio.dialogs.create_text_input_dialog(
        title="Hello",
        label="Enter your name:",
    )


def create_output() -> qtrio.dialogs.MessageBox:
    return qtrio.dialogs.create_message_box(
        title="Hello",
        text="",
        icon=QtWidgets.QMessageBox.Icon.Question,
        buttons=QtWidgets.QMessageBox.Ok,
    )


async def main(
    input_dialog: typing.Optional[qtrio.dialogs.TextInputDialog] = None,
    output_dialog: typing.Optional[qtrio.dialogs.MessageBox] = None,
) -> None:
    if input_dialog is None:  # pragma: no cover
        input_dialog = create_input()

    if output_dialog is None:  # pragma: no cover
        output_dialog = create_output()

    with contextlib.suppress(qtrio.UserCancelledError):
        name = await input_dialog.wait()

        output_dialog.text = f"Hi {name}, welcome to the team!"

        await output_dialog.wait()


if __name__ == "__main__":  # pragma: no cover
    qtrio.run(main)



import typing

import qtrio
from qtpy import QtCore
import trio
import trio.testing

import qtrio.examples.emissions


async def test_main(qtbot):
    window = qtrio.examples.emissions.Window.build()
    qtbot.addWidget(window.widget)

    results: typing.List[str] = []

    async def user():
        await emissions.channel.receive()

        buttons = [
            window.increment,
            window.increment,
            window.increment,
            window.decrement,
            window.decrement,
            window.decrement,
            window.decrement,
        ]
        for button in buttons:
            # TODO: Doesn't work reliably on macOS in GitHub Actions.  Seems to
            #       sometimes just miss the click entirely.
            # qtbot.mouseClick(button, QtCore.Qt.LeftButton)
            button.click()
            await trio.testing.wait_all_tasks_blocked(cushion=0.01)
            results.append(window.label.text())

        window.widget.close()

    async with trio.open_nursery() as nursery:
        async with qtrio.enter_emissions_channel(
            signals=[window.widget.shown],
        ) as emissions:
            nursery.start_soon(user)

            await qtrio.examples.emissions.main(window=window)

    assert results == ["1", "2", "3", "2", "1", "0", "-1"]



import threading
import time
import typing

import outcome
import pytest
from qtpy import QtCore
import qtrio
import qtrio._core
import trio
import trio.testing


@pytest.fixture(
    name="emissions_channel",
    params=[
        qtrio._core.open_emissions_channel,
        qtrio.enter_emissions_channel,
    ],
    ids=["qtrio._core.open_emissions_channel", "qtrio.enter_emissions_channel"],
)
def emissions_channel_fixture(request):
    return request.param


def test_reenter_event_triggers_in_main_thread(qapp):
    """Reenter events posted in another thread result in the function being run in the
    main thread.
    """
    result = []

    reenter = qtrio._core.Reenter()

    def post():
        qtrio.register_event_type()
        event = qtrio._core.ReenterEvent(fn=handler)
        qapp.postEvent(reenter, event)

    def handler():
        result.append(threading.get_ident())

    thread = threading.Thread(target=post)
    thread.start()
    thread.join()

    qapp.processEvents()

    assert result == [threading.get_ident()]


timeout = 40


def test_run_returns_value(testdir):
    """:func:`qtrio.run()` returns the result of the passed async function."""

    test_file = r"""
    import outcome

    import qtrio

    def test():
        async def main():
            return 29

        result = qtrio.run(main)

        assert result == 29
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_run_passes_args(testdir):
    """:func:`qtrio.run()` passes *args to async function."""

    test_file = r"""
    import outcome

    import qtrio

    def test():
        result = []
        
        async def main(arg1, arg2):
            result.append(arg1)
            result.append(arg2)

        qtrio.run(main, 27, 32)

        assert result == [27, 32]
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_qt_last_window_closed_does_not_quit_qt_or_cancel_trio(testdir):
    """When the last Qt window is closed, application and Trio both continue."""

    test_file = r"""
    import outcome
    from qtpy import QtCore
    import trio

    import qtrio


    def test():
        async def main():
            counter = 0

            def f():
                nonlocal counter
                application.lastWindowClosed.emit()
                counter = 1

            application = QtCore.QCoreApplication.instance()
            QtCore.QTimer.singleShot(100, f)

            while True:
                await trio.sleep(0.1)
                if counter == 1:
                    counter += 1
                elif counter > 1:
                    return counter

        result = qtrio.run(async_fn=main)

        assert result == 2
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_qt_quit_cancels_trio_with_custom_application(testdir):
    """When a passed Qt application exits the main Trio function is cancelled."""

    test_file = r"""
    import outcome
    from qtpy import QtCore
    from qtpy import QtWidgets
    import trio

    import qtrio


    def test():
        async def main():
            QtCore.QTimer.singleShot(
                100,
                QtCore.QCoreApplication.instance().lastWindowClosed.emit,
            )

            while True:
                await trio.sleep(1)

        runner = qtrio.Runner(application=QtWidgets.QApplication([]))
        outcomes = runner.run(async_fn=main)

        assert outcomes.trio.value == None
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_run_passes_internal_too_slow_error(testdir):
    """The async function run by :func:`qtrio.run` is executed in the Qt host thread."""

    test_file = r"""
    import math
    import pytest

    import qtrio
    import trio


    def test():
        async def main():
            with trio.fail_after(0):
                await trio.sleep(math.inf)

        with pytest.raises(trio.TooSlowError):
            qtrio.run(main)
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_run_runs_in_main_thread(testdir):
    """The async function run by :func:`qtrio.run` is executed in the Qt host thread."""

    test_file = r"""
    import threading

    import qtrio


    def test():
        async def main():
            return threading.get_ident()

        result = qtrio.run(main)

        assert result == threading.get_ident()
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_runner_runs_in_main_thread(testdir):
    """A directly used :class:`qtrio.Runner` runs the async function in the Qt host
    thread.
    """
    test_file = r"""
    import threading

    import qtrio


    def test():
        async def main():
            return threading.get_ident()

        runner = qtrio.Runner()
        outcomes = runner.run(main)

        assert outcomes.trio.value == threading.get_ident()
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_done_callback_runs_in_main_thread(testdir):
    """The done callback run by the Trio guest when finished is run in the Qt host
    thread.
    """
    test_file = r"""
    import threading

    import qtrio


    def test():
        result = {}

        async def main():
            pass

        def done_callback(outcomes):
            result['thread_id'] = threading.get_ident()

        qtrio.run(async_fn=main, done_callback=done_callback)

        assert result['thread_id'] == threading.get_ident()
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_done_callback_gets_outcomes(testdir):
    """The done callback is passed the outcomes with the Trio entry filled."""
    test_file = r"""
    import outcome

    import qtrio


    def test():
        result = {}

        async def main():
            return 93

        def done_callback(outcomes):
            result['outcomes'] = outcomes

        qtrio.run(async_fn=main, done_callback=done_callback)

        assert result['outcomes'] == qtrio.Outcomes(
            qt=None,
            trio=outcome.Value(93),
        )
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_out_of_hints_raises(testdir):
    """If there are no available Qt event types remaining
    EventTypeRegistrationFailedError is raised when requesting a type without any
    specific requested value.
    """
    test_file = r"""
    import pytest
    from qtpy import QtCore
    import qtrio


    def test():
        while QtCore.QEvent.registerEventType() != -1:
            # use up all the event types
            pass

        with pytest.raises(qtrio.EventTypeRegistrationFailedError):
            qtrio.register_event_type()
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_out_of_hints_raises_for_requested(testdir):
    """If there are no available Qt event types remaining
    EventTypeRegistrationFailedError is raised when requesting a type with a specific
    requested value.
    """
    test_file = r"""
    import pytest
    from qtpy import QtCore
    import qtrio


    def test():
        while QtCore.QEvent.registerEventType() != -1:
            # use up all the event types
            pass

        with pytest.raises(qtrio.EventTypeRegistrationFailedError):
            qtrio.register_requested_event_type(QtCore.QEvent.User)
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_out_of_hints_raises_when_requesting_already_used_type(testdir):
    """An error is raised when a request is made to register an already used event
    type.
    """
    test_file = r"""
    import pytest
    from qtpy import QtCore
    import qtrio


    def test():
        already_used = QtCore.QEvent.registerEventType()

        with pytest.raises(qtrio.RequestedEventTypeUnavailableError):
            qtrio.register_requested_event_type(already_used)
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_requesting_available_event_type_succeeds(testdir):
    """Requesting an available event type succeeds."""
    test_file = r"""
    from qtpy import QtCore
    import qtrio


    def test():
        qtrio.register_requested_event_type(QtCore.QEvent.User)

        assert qtrio.registered_event_type() == QtCore.QEvent.User
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_registering_event_type_when_already_registered(testdir):
    """Requesting an available event type succeeds."""
    test_file = r"""
    import pytest
    import qtrio


    def test():
        qtrio.register_event_type()

        with pytest.raises(qtrio.EventTypeAlreadyRegisteredError):
            qtrio.register_event_type()
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_registering_requested_event_type_when_already_registered(testdir):
    """Requesting an available event type succeeds."""
    test_file = r"""
    import pytest
    import qtrio


    def test():
        qtrio.register_event_type()

        with pytest.raises(qtrio.EventTypeAlreadyRegisteredError):
            qtrio.register_requested_event_type(qtrio.registered_event_type())
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


async def test_wait_signal_waits():
    """wait_signal() waits for the signal."""

    timer = QtCore.QTimer()
    timer.setSingleShot(True)

    start = time.monotonic()

    timer.start(100)

    await qtrio._core.wait_signal(timer.timeout)

    end = time.monotonic()

    assert end - start > 0.090


async def test_wait_signal_returns_the_value():
    """wait_signal() waits for the signal."""

    class MyQObject(QtCore.QObject):
        signal = QtCore.Signal(int)

    async def emit(signal, value):
        signal.emit(value)

    instance = MyQObject()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(emit, instance.signal, 17)
        result = await qtrio._core.wait_signal(instance.signal)

    assert result == (17,)


async def test_wait_signal_context_waits():
    """wait_signal_context() waits for the signal."""

    timer = QtCore.QTimer()
    timer.setSingleShot(True)

    async with qtrio._core.wait_signal_context(signal=timer.timeout):
        start = time.monotonic()
        timer.start(100)

    end = time.monotonic()

    assert end - start > 0.090


def test_outcomes_unwrap_none():
    """Unwrapping an empty Outcomes raises NoOutcomesError."""
    this_outcome = qtrio.Outcomes()

    with pytest.raises(qtrio.NoOutcomesError):
        this_outcome.unwrap()


def test_outcomes_unwrap_returns_trio_value_over_qt_none():
    """Unwrapping an Outcomes prioritizes a Trio value over a Qt None."""
    this_outcome = qtrio.Outcomes(trio=outcome.Value(3))
    result = this_outcome.unwrap()

    assert result == 3


def test_outcomes_unwrap_returns_trio_value_over_qt_value():
    """Unwrapping an Outcomes prioritizes a Trio value over a Qt value."""
    this_outcome = qtrio.Outcomes(qt=outcome.Value(2), trio=outcome.Value(3))
    result = this_outcome.unwrap()

    assert result == 3


def test_outcomes_unwrap_raises_trio_error_over_qt_none():
    """Unwrapping an Outcomes prioritizes a Trio error over a Qt None."""

    class LocalUniqueException(Exception):
        pass

    this_outcome = qtrio.Outcomes(trio=outcome.Error(LocalUniqueException()))
    with pytest.raises(LocalUniqueException):
        this_outcome.unwrap()


def test_outcomes_unwrap_raises_trio_error_over_qt_value():
    """Unwrapping an Outcomes prioritizes a Trio error over a Qt value."""

    class LocalUniqueException(Exception):
        pass

    this_outcome = qtrio.Outcomes(
        qt=outcome.Value(9),
        trio=outcome.Error(LocalUniqueException()),
    )

    with pytest.raises(LocalUniqueException):
        this_outcome.unwrap()


def test_outcomes_unwrap_raises_trio_error_over_qt_error():
    """Unwrapping an Outcomes prioritizes a Trio error over a Qt error."""

    class LocalUniqueException(Exception):
        pass

    class AnotherUniqueException(Exception):
        pass

    this_outcome = qtrio.Outcomes(
        qt=outcome.Error(AnotherUniqueException()),
        trio=outcome.Error(LocalUniqueException()),
    )

    with pytest.raises(LocalUniqueException):
        this_outcome.unwrap()


def test_outcomes_unwrap_returns_qt_value_over_trio_none():
    """Unwrapping an Outcomes prioritizes a Qt value over a Trio None."""
    this_outcome = qtrio.Outcomes(qt=outcome.Value(3))
    result = this_outcome.unwrap()

    assert result == 3


def test_outcomes_unwrap_raises_qt_error_over_trio_none():
    """Unwrapping an Outcomes prioritizes a Qt error over a Trio None."""

    class LocalUniqueException(Exception):
        pass

    this_outcome = qtrio.Outcomes(qt=outcome.Error(LocalUniqueException()))
    with pytest.raises(LocalUniqueException):
        this_outcome.unwrap()


def test_outcomes_unwrap_raises_qt_error_over_trio_value():
    """Unwrapping an Outcomes prioritizes a Qt error over a Trio value."""

    class LocalUniqueException(Exception):
        pass

    this_outcome = qtrio.Outcomes(
        qt=outcome.Error(LocalUniqueException()),
        trio=outcome.Value(8),
    )

    with pytest.raises(LocalUniqueException):
        this_outcome.unwrap()


def test_outcome_from_application_return_code_value():
    """Zero return code results in outcome.Value."""
    result = qtrio._core.outcome_from_application_return_code(return_code=0)

    assert result == outcome.Value(0)


def test_outcome_from_application_return_code_error():
    """Non-zero return code result in outcome.Error"""
    result = qtrio._core.outcome_from_application_return_code(return_code=-1)

    assert result == outcome.Error(qtrio.ReturnCodeError(-1))


def test_failed_hosted_trio_exception_on_stdout(testdir):
    """Except is printed when main Trio function raises."""

    test_file = r"""
    from qtpy import QtCore
    import qtrio


    class UniqueLocalException(Exception):
        pass


    async def test():
        raise UniqueLocalException("with a message")
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest()
    result.assert_outcomes(failed=1)
    result.stdout.fnmatch_lines(
        lines2=[
            "*test_failed_hosted_trio_exception_on_stdout.UniqueLocalException: with a message*"
        ]
    )


def test_emissions_equal():
    """:class:`Emission` objects created from the same :class:`QtCore.Signal` instance
    and args are equal even if the attributes are different instances.
    """

    class C(QtCore.QObject):
        signal = QtCore.Signal()

    instance = C()

    assert qtrio._core.Emission(
        signal=instance.signal, args=(13,)
    ) == qtrio._core.Emission(signal=instance.signal, args=(13,))


def test_emissions_unequal_by_signal():
    """:class:`Emission` objects with the same arguments but different signals are
    unequal.
    """

    class C(QtCore.QObject):
        signal_a = QtCore.Signal()
        signal_b = QtCore.Signal()

    instance = C()

    assert qtrio._core.Emission(
        signal=instance.signal_a, args=(13,)
    ) != qtrio._core.Emission(signal=instance.signal_b, args=(13,))


def test_emissions_unequal_by_instance():
    """:class:`Emission` objects with the same signal but on different instances are
    unequal.
    """

    class C(QtCore.QObject):
        signal = QtCore.Signal()

    instance_a = C()
    instance_b = C()

    assert qtrio._core.Emission(
        signal=instance_a.signal, args=(13,)
    ) != qtrio._core.Emission(signal=instance_b.signal, args=(13,))


def test_emissions_unequal_by_type():
    """:class:`Emission` objects are not equal to integers"""

    class C(QtCore.QObject):
        signal = QtCore.Signal()

    instance = C()

    assert qtrio._core.Emission(signal=instance.signal, args=(13,)) != 13


def test_emissions_unequal_by_args():
    """:class:`Emission` objects with the same signal but different arguments are
    unequal.
    """

    class C(QtCore.QObject):
        signal = QtCore.Signal()

    instance = C()

    assert qtrio._core.Emission(
        signal=instance.signal, args=(13,)
    ) != qtrio._core.Emission(signal=instance.signal, args=(14,))


async def test_emissions_channel_iterates_one(emissions_channel):
    """Emissions channel yields one emission as expected."""

    class MyQObject(QtCore.QObject):
        signal = QtCore.Signal(int)

    instance = MyQObject()

    async with emissions_channel(signals=[instance.signal]) as emissions:
        instance.signal.emit(93)
        await trio.testing.wait_all_tasks_blocked(cushion=0.01)
        await emissions.aclose()

        async with emissions.channel:
            emissions = [result async for result in emissions.channel]

    assert emissions == [qtrio._core.Emission(signal=instance.signal, args=(93,))]


async def test_emissions_channel_iterates_three(emissions_channel):
    """Emissions channel yields three emissions as expected."""

    class MyQObject(QtCore.QObject):
        signal = QtCore.Signal(int)

    instance = MyQObject()

    async with emissions_channel(signals=[instance.signal]) as emissions:
        for v in [93, 56, 27]:
            instance.signal.emit(v)
            await trio.testing.wait_all_tasks_blocked(cushion=0.01)
        await emissions.aclose()

        async with emissions.channel:
            emissions = [result async for result in emissions.channel]

    assert emissions == [
        qtrio._core.Emission(signal=instance.signal, args=(v,)) for v in [93, 56, 27]
    ]


async def test_emissions_channel_with_three_receives_first(emissions_channel):
    """Emissions channel yields receives first item when requested."""

    class MyQObject(QtCore.QObject):
        signal = QtCore.Signal(int)

    instance = MyQObject()

    async with emissions_channel(signals=[instance.signal]) as emissions:
        for v in [93, 56, 27]:
            instance.signal.emit(v)
            await trio.testing.wait_all_tasks_blocked(cushion=0.01)
        await emissions.aclose()

        async with emissions.channel:
            emission = await emissions.channel.receive()

    assert emission == qtrio._core.Emission(signal=instance.signal, args=(93,))


async def test_emissions_channel_iterates_in_order(emissions_channel):
    """Emissions channel yields signal emissions in order (pretty probably...)."""

    class MyQObject(QtCore.QObject):
        signal = QtCore.Signal(int)

    instance = MyQObject()
    values = list(range(100))
    results = []

    async with emissions_channel(
        signals=[instance.signal],
        max_buffer_size=len(values),
    ) as emissions:
        for i, v in enumerate(values):
            if i % 2 == 0:
                await trio.testing.wait_all_tasks_blocked(cushion=0.001)

            instance.signal.emit(v)

        await emissions.aclose()

        async with emissions.channel:
            async for emission in emissions.channel:
                [value] = emission.args
                results.append(value)

    assert results == values


async def test_emissions_channel_limited_buffer(emissions_channel):
    """Emissions channel throws away beyond buffer limit."""

    class MyQObject(QtCore.QObject):
        signal = QtCore.Signal(int)

    instance = MyQObject()
    max_buffer_size = 10
    values = list(range(2 * max_buffer_size))
    results = []

    async with emissions_channel(
        signals=[instance.signal],
        max_buffer_size=max_buffer_size,
    ) as emissions:
        for v in values:
            instance.signal.emit(v)

        await emissions.aclose()

        async with emissions.channel:
            async for emission in emissions.channel:
                [value] = emission.args
                results.append(value)

    assert results == values[:max_buffer_size]


async def test_open_emissions_channel_does_not_close_read_channel():
    """Exiting open_emissions_channel() closes send channel and does not close
    read channel on exit.
    """

    class MyQObject(QtCore.QObject):
        signal = QtCore.Signal(int)

    instance = MyQObject()
    max_buffer_size = 10

    async with qtrio._core.open_emissions_channel(
        signals=[instance.signal],
        max_buffer_size=max_buffer_size,
    ) as emissions:
        pass

    with pytest.raises(trio.EndOfChannel):
        emissions.channel.receive_nowait()

    with pytest.raises(trio.ClosedResourceError):
        emissions.send_channel.send_nowait(None)


async def test_enter_emissions_channel_closes_both_channels():
    """Exiting enter_emissions_channel() closes send and receive channels on exit."""

    class MyQObject(QtCore.QObject):
        signal = QtCore.Signal(int)

    instance = MyQObject()
    max_buffer_size = 10

    async with qtrio.enter_emissions_channel(
        signals=[instance.signal],
        max_buffer_size=max_buffer_size,
    ) as emissions:
        pass

    with pytest.raises(trio.ClosedResourceError):
        emissions.channel.receive_nowait()

    with pytest.raises(trio.ClosedResourceError):
        emissions.send_channel.send_nowait(None)


def emissions_nursery_connect_maybe_async(
    is_async: bool,
    nursery: qtrio.EmissionsNursery,
    signal: qtrio._util.SignalInstance,
    slot: typing.Callable[..., object],
) -> None:
    if is_async:

        async def async_slot(*args):
            return slot(*args)

        nursery.connect(signal, async_slot)
    else:
        nursery.connect_sync(signal, slot)


@pytest.fixture(name="is_async", params=[False, True], ids=["sync", "async"])
def is_async_fixture(request):
    yield request.param


async def test_emissions_nursery_runs_callbacks(is_async):
    """Callbacks connected to an emissions nursery get run."""

    class SignalHost(QtCore.QObject):
        signal = QtCore.Signal(int)

    results = set()
    event = trio.Event()

    def slot(number):
        results.add(number)
        if len(results) == 5:
            event.set()

    async with qtrio.open_emissions_nursery() as emissions_nursery:
        signal_host = SignalHost()

        emissions_nursery_connect_maybe_async(
            is_async=is_async,
            nursery=emissions_nursery,
            signal=signal_host.signal,
            slot=slot,
        )

        for i in range(5):
            signal_host.signal.emit(i)

        await event.wait()

    assert results == {0, 1, 2, 3, 4}


async def test_emissions_nursery_disconnects(is_async):
    """Callbacks are disconnected when exiting the context and aren't run for emissions
    after leaving.
    """

    class SignalHost(QtCore.QObject):
        signal = QtCore.Signal(int)

    results = set()

    def slot(number):
        results.add(number)  # pragma: no cover

    async with qtrio.open_emissions_nursery() as emissions_nursery:
        signal_host = SignalHost()

        emissions_nursery_connect_maybe_async(
            is_async=is_async,
            nursery=emissions_nursery,
            signal=signal_host.signal,
            slot=slot,
        )

    for i in range(5):
        signal_host.signal.emit(i)

    await trio.testing.wait_all_tasks_blocked(cushion=0.01)

    assert results == set()


async def test_emissions_nursery_cancellation_cancels_callbacks():
    """Callbacks are cancelled when the nursery is cancelled."""

    class SignalHost(QtCore.QObject):
        signal = QtCore.Signal(int)

    event = trio.Event()
    results = set()
    waiting = set()

    async def slot(number):
        try:
            waiting.add(number)

            if len(waiting) == 5:
                event.set()

            while True:
                await trio.sleep(1)
        except trio.Cancelled:
            results.add(number)
            raise

    async with qtrio.open_emissions_nursery() as emissions_nursery:
        signal_host = SignalHost()

        emissions_nursery.connect(signal_host.signal, slot)

        for i in range(5):
            signal_host.signal.emit(i)

        await event.wait()
        emissions_nursery.nursery.cancel_scope.cancel()

    assert results == {0, 1, 2, 3, 4}


async def test_emissions_nursery_receives_exceptions(is_async):
    """Callbacks that raise exceptions will feed them out to the nursery."""

    class SignalHost(QtCore.QObject):
        signal = QtCore.Signal()

    class LocalUniqueException(Exception):
        pass

    def slot():
        raise LocalUniqueException()

    with pytest.raises(LocalUniqueException):
        async with qtrio.open_emissions_nursery() as emissions_nursery:
            signal_host = SignalHost()

            emissions_nursery_connect_maybe_async(
                is_async=is_async,
                nursery=emissions_nursery,
                signal=signal_host.signal,
                slot=slot,
            )

            signal_host.signal.emit()


async def test_emissions_nursery_waits_for_until_signal():
    """Emissions nursery waits to exit until `until` signal is emitted."""

    class SignalHost(QtCore.QObject):
        signal = QtCore.Signal()

    results = []

    signal_host = SignalHost()

    async def emit_later():
        results.append(1)
        await trio.testing.wait_all_tasks_blocked(cushion=0.01)
        results.append(2)
        signal_host.signal.emit()

    async with trio.open_nursery() as nursery:
        async with qtrio.open_emissions_nursery(until=signal_host.signal):
            nursery.start_soon(emit_later)
            results.append(0)

        results.append(3)

    assert results == [0, 1, 2, 3]


async def test_emissions_nursery_wraps(is_async):
    """Emissions nursery wraps callbacks as requested."""

    class SignalHost(QtCore.QObject):
        signal = QtCore.Signal()

    class LocalUniqueException(Exception):
        pass

    result: outcome.Outcome

    event = trio.Event()
    signal_host = SignalHost()

    async def wrapper(asyncfn, *args):
        nonlocal result

        try:
            await asyncfn(*args)
        except Exception as e:
            result = outcome.Error(e)
            event.set()

    def slot():
        raise LocalUniqueException()

    async with qtrio.open_emissions_nursery(wrapper=wrapper) as emissions_nursery:
        emissions_nursery_connect_maybe_async(
            is_async=is_async,
            nursery=emissions_nursery,
            signal=signal_host.signal,
            slot=slot,
        )

        signal_host.signal.emit()
        await event.wait()

    with pytest.raises(LocalUniqueException):
        result.unwrap()


def test_run_without_executing_application(testdir):
    """Running without executing the application...  doesn't."""

    """Exiting enter_emissions_channel() closes send and receive channels on exit."""
    test_file = r"""
    import qtrio


    def test():
        ran = False

        async def async_fn():
            nonlocal ran
            ran = True

        runner = qtrio.Runner()
        runner.run(async_fn=async_fn, execute_application=False)

        assert not ran
    """

    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_execute_manually(testdir):
    """Executing manually works."""

    test_file = r"""
    import qtrio


    def test():
        ran = False

        async def async_fn():
            nonlocal ran
            ran = True

        runner = qtrio.Runner()
        runner.run(async_fn=async_fn, execute_application=False)

        assert not ran

        runner.application.exec_()

        assert ran
    """

    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)


def test_not_quitting_application_does_not(testdir):
    """Not quitting the application doesn't quit."""

    test_file = r"""
    from qtpy import QtCore
    import qtrio


    def test(qapp):
        results = []

        async def async_fn():
            pass

        def cycle():
            if runner._done and 'done' not in results:
                results.append('done')

                timer.stop()
                runner.application.quit()

        def about_to_quit():
            results.append('about to quit')

        timer = QtCore.QTimer()
        timer.setInterval(10)
        timer.timeout.connect(cycle)
        timer.start()

        runner = qtrio.Runner(application=qapp, quit_application=False)
        runner.application.aboutToQuit.connect(about_to_quit)
        runner.run(async_fn=async_fn)

        assert results == ['done', 'about to quit']
    """

    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess(timeout=timeout)
    result.assert_outcomes(passed=1)



import os
import pathlib
import subprocess
import sys
import sysconfig


def test_emissions_help_entry_point():
    """The console script entry point doesn't fail when asked for --help for the
    emissions example.
    """
    scripts_string = sysconfig.get_path("scripts")

    if scripts_string is None:  # pragma: no cover
        assert False, "No scripts path found in sysconfig paths"

    scripts = pathlib.Path(scripts_string)
    subprocess.run(
        [os.fspath(scripts / "qtrio"), "examples", "emissions", "--help"], check=True
    )


def test_emissions_help_dash_m():
    """The CLI run via ``python -m qtrio`` doesn't fail when asked for --help for the
    emissions example.
    """
    subprocess.run(
        [sys.executable, "-m", "qtrio", "examples", "emissions", "--help"], check=True
    )



import typing

import qtrio
from qtpy import QtWidgets
import trio


async def main(
    label: typing.Optional[QtWidgets.QWidget] = None,
    message: str = "Hello world.",
    change_delay: float = 0.5,
    close_delay: float = 3,
) -> None:
    if label is None:  # pragma: no cover
        label = QtWidgets.QLabel()
    # start big enough to fit the whole message
    label.setText(message)
    label.show()
    label.setText("")

    for i in range(len(message)):
        await trio.sleep(change_delay)
        label.setText(message[: i + 1])

    await trio.sleep(close_delay)


if __name__ == "__main__":  # pragma: no cover
    qtrio.run(main)



import typing

import qtpy
from qtpy import QtCore


SignalInstance: typing.Any
if qtpy.API in qtpy.PYQT5_API and not hasattr(QtCore, "SignalInstance"):
    SignalInstance = QtCore.pyqtBoundSignal
else:
    SignalInstance = QtCore.SignalInstance


T = typing.TypeVar("T")


class ProtocolChecker(typing.Generic[T]):
    """Instances of this class can be used as decorators that will result in type hint
    checks to verifying that other classes implement a given protocol.  Generally you
    would create a single instance where you define each protocol and then use that
    instance as the decorator.  Note that this usage is, at least in part, due to
    Python not supporting type parameter specification in the ``@`` decorator
    expression.

    .. code-block:: python

       import typing


       class MyProtocol(typing.Protocol):
           def a_method(self): ...

       check_my_protocol = qtrio._util.ProtocolChecker[MyProtocol]()

       @check_my_protocol
       class AClass:
           def a_method(self):
               return 42092
    """

    def __call__(self, cls: typing.Type[T]) -> typing.Type[T]:
        return cls



import sys
import typing


def main(
    input_file: typing.TextIO = sys.stdin, output_file: typing.TextIO = sys.stdout
) -> None:
    try:
        output_file.write("What is your name? ")
        output_file.flush()
        name = input_file.readline()[:-1]
        output_file.write(f"Hi {name}, welcome to the team!\n")
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":  # pragma: no cover
    main()



import typing

import attr
import qtrio
from qtpy import QtWidgets


@attr.s(auto_attribs=True)
class Widget:
    widget: QtWidgets.QWidget = attr.ib(factory=QtWidgets.QWidget)
    layout: QtWidgets.QLayout = attr.ib(factory=QtWidgets.QVBoxLayout)
    button: QtWidgets.QPushButton = attr.ib(factory=QtWidgets.QPushButton)
    label: QtWidgets.QWidget = attr.ib(factory=QtWidgets.QLabel)

    def setup(self, message: str) -> None:
        self.button.setText("More")

        # start big enough to fit the whole message
        self.label.setText(message)

        self.layout.addWidget(self.button)
        self.layout.addWidget(self.label)
        self.widget.setLayout(self.layout)

    def show(self) -> None:
        self.widget.show()
        self.label.setText("")


async def main(
    widget: typing.Optional[Widget] = None,
    message: str = "Hello world.",
) -> None:
    if widget is None:  # pragma: no cover
        widget = Widget()

    widget.setup(message=message)

    async with qtrio.enter_emissions_channel(
        signals=[widget.button.clicked]
    ) as emissions:
        i = 1
        widget.show()

        async for _ in emissions.channel:  # pragma: no branch
            widget.label.setText(message[:i])
            i += 1

            if i > len(message):
                break

        # wait for another click to finish
        await emissions.channel.receive()


if __name__ == "__main__":  # pragma: no cover
    qtrio.run(main)



"""Top-level package for QTrio."""

from ._version import __version__

from ._exceptions import (
    QTrioException,
    NoOutcomesError,
    EventTypeRegistrationError,
    EventTypeRegistrationFailedError,
    RequestedEventTypeUnavailableError,
    EventTypeAlreadyRegisteredError,
    ReturnCodeError,
    RunnerTimedOutError,
    UserCancelledError,
    InvalidInputError,
    InternalError,
    DialogNotActiveError,
)

from ._core import (
    enter_emissions_channel,
    open_emissions_nursery,
    Emissions,
    Emission,
    EmissionsNursery,
    Outcomes,
    run,
    Runner,
    registered_event_type,
    register_event_type,
    register_requested_event_type,
    Reenter,
)

from ._qt import Signal



from qtpy import QtCore
import pytest

import qtrio._qt


def test_signal_emits(qtbot):
    """qtrio._core.Signal emits."""

    class NotQObject:
        signal = qtrio.Signal()

    instance = NotQObject()

    with qtbot.wait_signal(instance.signal, 100):
        instance.signal.emit()


def test_signal_emits_value(qtbot):
    """qtrio._core.Signal emits a value."""

    class NotQObject:
        signal = qtrio.Signal(int)

    result = None

    def collect_result(value):
        nonlocal result
        result = value

    instance = NotQObject()
    instance.signal.connect(collect_result)

    with qtbot.wait_signal(instance.signal, 100):
        instance.signal.emit(13)

    assert result == 13


def test_accessing_signal_on_class_results_in_our_signal():
    """qtrio._core.Signal instance accessible via class attribute."""

    class NotQObject:
        signal = qtrio.Signal(int)

    assert isinstance(NotQObject.signal, qtrio.Signal)


def test_our_signal_object_method_returns_qobject():
    """qtrio._core.Signal instance provides access to signal-hosting QObject."""

    class NotQObject:
        signal = qtrio.Signal(int)

    instance = NotQObject()

    assert isinstance(NotQObject.signal.object(instance=instance), QtCore.QObject)


def test_connection_connects(qtbot):
    """qtrio._core.connection connects signal inside managed context."""

    class MyQObject(QtCore.QObject):
        signal = QtCore.Signal(int)

    instance = MyQObject()

    results = []

    def collect_result(value):
        results.append(value)

    instance.signal.emit(1)

    with qtrio._qt.connection(instance.signal, collect_result):
        instance.signal.emit(2)

    assert results == [2]


def test_connection_disconnects(qtbot):
    """qtrio._core.connection disconnects signal when exiting managed context."""

    class MyQObject(QtCore.QObject):
        signal = QtCore.Signal(int)

    instance = MyQObject()

    results = []

    def collect_result(value):
        results.append(value)

    with qtrio._qt.connection(instance.signal, collect_result):
        instance.signal.emit(1)

    instance.signal.emit(2)

    assert results == [1]


def test_connection_yield_can_be_disconnected(qtbot):
    """qtrio._core.connection result can be used to disconnect the signal early."""

    class MyQObject(QtCore.QObject):
        signal = QtCore.Signal(int)

    instance = MyQObject()

    results = []

    def collect_result(value):
        results.append(value)

    with qtrio._qt.connection(instance.signal, collect_result) as connection:
        instance.signal.emit(1)
        instance.signal.disconnect(connection)
        instance.signal.emit(2)

    assert results == [1]


def test_failed_connection_raises():
    """qtrio._core.connection raises TypeError on failure to connect."""

    class MyQObject(QtCore.QObject):
        signal = QtCore.Signal()

    instance = MyQObject()

    # TODO: get more specific about the exception
    with pytest.raises(TypeError):
        with qtrio._qt.connection(instance.signal, 2):  # type: ignore
            pass  # pragma: no cover



import pytest
from qtpy import QtWidgets


@pytest.fixture(name="qtrio_preshow_workaround", scope="session", autouse=True)
def qtrio_preshow_workaround_fixture(qapp):
    dialog = QtWidgets.QMessageBox(
        QtWidgets.QMessageBox.Information,
        "",
        "",
        QtWidgets.QMessageBox.Ok,
    )

    dialog.show()
    dialog.hide()


@pytest.fixture(name="qtrio_testdir", autouse=True)
def qtrio_testdir_fixture(testdir):
    text = """
    [pytest]
    trio_mode = true
    trio_run = trio
    """
    testdir.makefile(".ini", pytest=text)

    return testdir



import qtrio


def test_return_code_error_equality():
    assert qtrio.ReturnCodeError(1) == qtrio.ReturnCodeError(1)


def test_return_code_error_inequality_by_value():
    assert qtrio.ReturnCodeError(1) != qtrio.ReturnCodeError(2)


def test_return_code_error_inequality_by_type():
    assert qtrio.ReturnCodeError(1) != Exception()



import qtrio._tests.helpers

pytest_plugins = "pytester"

qtrio_preshow_workaround_fixture = qtrio._tests.helpers.qtrio_preshow_workaround_fixture
qtrio_testdir_fixture = qtrio._tests.helpers.qtrio_testdir_fixture



import os
import pathlib
import random
import sys
import typing

import hyperlink
import pytest
import trio

import qtrio.dialogs
import qtrio.examples.download


minimum_python_version_for_quart_trio = (3, 7)
if sys.version_info < minimum_python_version_for_quart_trio:
    minimum_python_version_string = ".".join(
        str(v) for v in minimum_python_version_for_quart_trio
    )
    python_version_string = ".".join(str(v) for v in sys.version_info)

    reason = (
        f"quart-trio is not available for Python <{minimum_python_version_string}."
        f"  Running in Python {python_version_string}."
    )

    # quart-trio is expected to not be available but this is the public interface that
    # pytest provides for skipping and also terminating the import of the test module.
    pytest.importorskip(modname="quart_trio", reason=reason)
    raise Exception("this should never run")  # pragma: no cover


# By the time we get here, we know that quart-trio is supposed to be available.
import quart_trio


T = typing.TypeVar("T")


async def asynchronize(sequence: typing.Iterable[T]) -> typing.AsyncIterable[T]:
    for element in sequence:
        yield element


def randomly_chunked_bytes(
    random_generator: random.Random,
    chunk_size_minimum: int,
    chunk_size_maximum: int,
    chunk_count: int,
) -> typing.List[bytes]:
    return [
        bytes(
            random_generator.randrange(256)
            for _ in range(
                random_generator.randrange(chunk_size_minimum, chunk_size_maximum)
            )
        )
        for _ in range(chunk_count)
    ]


# Just using a seeded random.Random as an easy way to get consistent data without
# storing it all in the repository.
random_generator = random.Random(0)
one_chunk: typing.List[bytes] = [
    bytes(random_generator.randrange(256) for _ in range(1_000))
]
many_chunks: typing.List[bytes] = [
    *randomly_chunked_bytes(
        random_generator=random_generator,
        chunk_size_minimum=10,
        chunk_size_maximum=20,
        chunk_count=10,
    ),
    b"",
] * 2


@pytest.fixture(
    name="chunked_data",
    params=[
        many_chunks,
        one_chunk,
    ],
    ids=["Many chunks", "One chunk"],
)
def chunked_data_fixture(request):
    return request.param


@pytest.fixture(
    name="content_length_specified",
    params=[False, True],
    ids=["Unknown content length", "Known content length"],
)
def content_length_specified_fixture(request):
    return request.param


@pytest.fixture(name="content_length")
def content_length_fixture(content_length_specified, chunked_data):
    content_length: typing.Optional[int]

    if content_length_specified:
        content_length = sum(len(chunk) for chunk in chunked_data)
    else:
        content_length = None

    return content_length


@pytest.fixture(name="url")
def url_fixture():
    return hyperlink.URL.from_text("http://test/")


@pytest.fixture(name="http_application")
def http_application_fixture(chunked_data, content_length_specified, content_length):
    application = quart_trio.QuartTrio(__name__)

    @application.route("/")
    async def root():
        headers = {}

        if content_length_specified:
            headers["content-length"] = content_length

        return asynchronize(chunked_data), 200, headers

    return application


@pytest.fixture(
    name="optional_text_input_dialog",
    params=[False, True],
    ids=["No text input dialog", "Text input dialog"],
)
def optional_text_input_dialog_fixture(request):
    if request.param:
        return qtrio.dialogs.create_text_input_dialog()

    return None


@pytest.fixture(
    name="optional_file_dialog",
    params=[False, True],
    ids=["No file dialog", "File dialog"],
)
def optional_file_dialog_fixture(request):
    if request.param:
        return qtrio.dialogs.create_file_save_dialog()

    return None


async def test_main(
    chunked_data: typing.List[bytes],
    http_application: quart_trio.QuartTrio,
    optional_text_input_dialog: typing.Optional[qtrio.dialogs.TextInputDialog],
    optional_file_dialog: typing.Optional[qtrio.dialogs.FileDialog],
    url: hyperlink.URL,
    tmp_path: pathlib.Path,
) -> None:
    temporary_directory = trio.Path(tmp_path)

    data = b"".join(chunked_data)
    destination = temporary_directory.joinpath("file")

    progress_dialog = qtrio.dialogs.create_progress_dialog()
    message_box = qtrio.dialogs.create_message_box()

    async def user():
        if optional_text_input_dialog is not None:
            emission = await emissions.channel.receive()
            assert emission.is_from(optional_text_input_dialog.shown)

            assert optional_text_input_dialog.line_edit is not None
            optional_text_input_dialog.line_edit.setText(url.to_text())

            assert optional_text_input_dialog.accept_button is not None
            optional_text_input_dialog.accept_button.click()

        if optional_file_dialog is not None:
            emission = await emissions.channel.receive()
            assert emission.is_from(optional_file_dialog.shown)

            assert optional_file_dialog.dialog is not None
            await optional_file_dialog.set_path(path=destination)

            assert optional_file_dialog.accept_button is not None
            optional_file_dialog.accept_button.click()

        emission = await emissions.channel.receive()
        assert emission.is_from(message_box.shown)

        assert message_box.accept_button is not None
        message_box.accept_button.click()

    signals = []

    if optional_text_input_dialog is not None:
        signals.append(optional_text_input_dialog.shown)

    if optional_file_dialog is not None:
        signals.append(optional_file_dialog.shown)

    signals.append(message_box.shown)

    async with qtrio.enter_emissions_channel(signals=signals) as emissions:
        async with trio.open_nursery() as nursery:
            nursery.start_soon(user)

            await qtrio.examples.download.main(
                url=url if optional_text_input_dialog is None else None,
                destination=destination if optional_file_dialog is None else None,
                fps=10,
                text_input_dialog=optional_text_input_dialog,
                file_dialog=optional_file_dialog,
                progress_dialog=progress_dialog,
                message_box=message_box,
                http_application=http_application,
            )

    async with await destination.open("rb") as written_file:
        written = await written_file.read()  # type: ignore[attr-defined]

    assert written == data


async def test_get_dialog(
    chunked_data: typing.List[bytes],
    http_application: quart_trio.QuartTrio,
    url: hyperlink.URL,
    tmp_path: pathlib.Path,
) -> None:
    temporary_directory = trio.Path(tmp_path)

    data = b"".join(chunked_data)
    destination = temporary_directory.joinpath("file")

    progress_dialog = qtrio.dialogs.create_progress_dialog()
    message_box = qtrio.dialogs.create_message_box()

    async def user():
        await emissions.channel.receive()

        assert message_box.accept_button is not None
        message_box.accept_button.click()

    async with qtrio.enter_emissions_channel(
        signals=[message_box.shown],
    ) as emissions:
        async with trio.open_nursery() as nursery:
            nursery.start_soon(user)

            await qtrio.examples.download.get_dialog(
                url=url,
                destination=destination,
                fps=0.1,
                progress_dialog=progress_dialog,
                message_box=message_box,
                http_application=http_application,
            )

    async with await destination.open("rb") as written_file:
        written = await written_file.read()  # type: ignore[attr-defined]

    assert written == data


async def test_get_dialog_canceled(
    chunked_data: typing.List[bytes],
    http_application: quart_trio.QuartTrio,
    url: hyperlink.URL,
    tmp_path: pathlib.Path,
) -> None:
    temporary_directory = trio.Path(tmp_path)

    destination = temporary_directory.joinpath("file")

    progress_dialog = qtrio.dialogs.create_progress_dialog()
    message_box = qtrio.dialogs.create_message_box()

    async def user():
        await emissions.channel.receive()

        assert progress_dialog.dialog is not None
        assert progress_dialog.dialog.isVisible()
        assert message_box.dialog is None

        assert progress_dialog.cancel_button is not None
        progress_dialog.cancel_button.click()

    async with qtrio.enter_emissions_channel(
        signals=[progress_dialog.shown],
    ) as emissions:
        with pytest.raises(qtrio.UserCancelledError):
            async with trio.open_nursery() as nursery:
                nursery.start_soon(user)

                await qtrio.examples.download.get_dialog(
                    url=url,
                    destination=destination,
                    fps=0.1,
                    progress_dialog=progress_dialog,
                    message_box=message_box,
                    http_application=http_application,
                )


async def test_get(
    chunked_data: typing.List[bytes],
    content_length: typing.Optional[int],
    http_application: quart_trio.QuartTrio,
    url: hyperlink.URL,
    tmp_path: pathlib.Path,
) -> None:
    temporary_directory = trio.Path(tmp_path)

    data = b"".join(chunked_data)
    destination = temporary_directory.joinpath("file")

    progresses = []

    async for progress in qtrio.examples.download.get(
        url=url,
        destination=destination,
        update_period=0,
        http_application=http_application,
    ):
        progresses.append(progress)

    async with await destination.open("rb") as written_file:
        written = await written_file.read()  # type: ignore[attr-defined]

    assert written == data
    assert all(progress.total == content_length for progress in progresses)
    assert [progresses[0].downloaded, progresses[-1].downloaded] == [0, len(data)]



from qtpy import QtWidgets

import qtrio.examples.readme.qt


def test_main(qtbot):
    input_dialog = qtrio.examples.readme.qt.create_input()
    output_dialog = qtrio.examples.readme.qt.create_output()

    qtbot.add_widget(input_dialog)
    qtbot.add_widget(output_dialog)

    text_to_enter = "everyone"

    main_object = qtrio.examples.readme.qt.Main(
        input_dialog=input_dialog,
        output_dialog=output_dialog,
    )

    main_object.setup()

    qtbot.wait_for_window_shown(input_dialog)

    [line_edit] = input_dialog.findChildren(QtWidgets.QLineEdit)
    line_edit.setText(text_to_enter)
    input_dialog.accept()

    qtbot.wait_for_window_shown(output_dialog)

    output_text = output_dialog.text()

    output_dialog.accept()

    assert text_to_enter in output_text


def test_main_cancelled(qtbot):
    input_dialog = qtrio.examples.readme.qt.create_input()
    output_dialog = qtrio.examples.readme.qt.create_output()

    qtbot.add_widget(input_dialog)
    qtbot.add_widget(output_dialog)

    text_to_enter = "everyone"

    main_object = qtrio.examples.readme.qt.Main(
        input_dialog=input_dialog,
        output_dialog=output_dialog,
    )

    main_object.setup()

    qtbot.wait_for_window_shown(input_dialog)

    [line_edit] = input_dialog.findChildren(QtWidgets.QLineEdit)
    line_edit.setText(text_to_enter)
    input_dialog.reject()

    # qtbot.wait(1_000)

    output_text = output_dialog.text()

    assert output_text == ""



import io
import typing

import qtrio.examples.readme.console


def test_normal_usage():
    input_file = io.StringIO("Ms. Console\n")
    output_file = io.StringIO()

    qtrio.examples.readme.console.main(input_file=input_file, output_file=output_file)

    assert (
        output_file.getvalue()
        == "What is your name? Hi Ms. Console, welcome to the team!\n"
    )


class CancellingStringIO(io.StringIO):
    def readline(  # type: ignore[override]
        self,
        size: typing.Optional[int] = -1,
    ) -> str:
        raise KeyboardInterrupt()


def test_cancellation():
    input_file = CancellingStringIO()
    output_file = io.StringIO()

    qtrio.examples.readme.console.main(input_file=input_file, output_file=output_file)

    assert output_file.getvalue() == "What is your name? "



import typing

import qtrio
from qtpy import QtCore
from qtpy import QtWidgets
import trio

import qtrio.examples.crossingpaths


async def test_main(qtbot):
    class SignaledLabel(QtWidgets.QLabel):
        text_changed = QtCore.Signal(str)

        def setText(self, *args, **kwargs):
            super().setText(*args, **kwargs)
            self.text_changed.emit(self.text())

    label = SignaledLabel()
    qtbot.addWidget(label)

    results: typing.List[str] = []

    async def user():
        async for emission in emissions.channel:
            [text] = emission.args
            results.append(text)

    async with trio.open_nursery() as nursery:
        async with qtrio.enter_emissions_channel(
            signals=[label.text_changed],
        ) as emissions:
            nursery.start_soon(user)

            await qtrio.examples.crossingpaths.main(
                label=label,
                message="test world",
                change_delay=0.01,
                close_delay=0.01,
            )

    assert results == [
        "test world",
        "",
        "t",
        "te",
        "tes",
        "test",
        "test ",
        "test w",
        "test wo",
        "test wor",
        "test worl",
        "test world",
    ]



import qtrio._python


def test_identifier_path():
    """Identifier path generation...  works?"""
    assert (
        qtrio._python.identifier_path(qtrio._python.identifier_path)
        == "__qtrio__python_identifier_path"
    )
